#!/usr/bin/env python
#
import unittest
import femagtools
try:
    import mock
except ImportError:
    import unittest.mock as mock
import os


class PropulationTest(unittest.TestCase):

    @mock.patch('femagtools.amazon.Engine._create_amazon_resource')
    def setUp(self, create_amazone_resource):
        create_amazone_resource = mock.Mock()
        self.engine = femagtools.amazon.Engine()

    def tearDown(self):
        self.engine = None

    def test_read_cloud_ini(self):
        testPath = os.path.join(os.path.split(__file__)[0], 'data')
        if len(testPath) == 0:
            testPath = os.path.join(os.path.abspath('.'), 'data')

        result = """Test Cloud init
export ACL=authenticated-read
export CLOUD_INIT={}/cloud_init.txt
export COMPANY_NAME=femag
export ENGINE=amazon
export FINISH_TASK_FILENAME=exit_code
export IMAGE_ID=ami-b0cc23df
export INSTANCE_TYPE=t2.micro
export SERVER_LOCATION=eu-central-1
export BUCKET_NAME=1
""".format(testPath)

        self.engine.config['CLOUD_INIT'] = "{}/cloud_init.txt".format(testPath)
        user_data = self.engine._read_cloud_init('1')
        self.assertEqual(user_data, result)

    # This test does not work, cause mock can not mock attributes in a class which are
    # not defined in the init method
    @mock.patch('femagtools.job.CloudJob')
    @mock.patch('femagtools.job.Task')
    def test_get_status_code(self, MockCloudJob, MockCloudTask):
        filename = "../data/exit_code"
        from femagtools.job import CloudTask

        # MockCloudJob.add_task = Mock()
        # MockCloudTask.directory.__get__ = Mock(return_value='/tmp')
        # MockCloudJob.tasks.return_value = [MockCloudTask, MockCloudTask, MockCloudTask]

        m = mock.MagicMock()
        mock.directory = mock.Mock(return_value='/tmp')
        self.engine.create_job('')

        self.engine.job.tasks.append(m(2, "/tmp/"))

        # This is the test
        # self.engine._get_status_code('exit_code')


if __name__ == '__main__':
    unittest.main()



import femagtools.config
import os.path
import platform


def test_femag():
    if platform.system() == "Windows":
        expected = "wfemagw64"
    else:
        expected = "xfemag64"

    try:
        msg = femagtools.config.get_executable()
    except Exception as e:
        msg = str(e)
    assert msg.find(expected) > -1


def test_engine_config():
    default_config = {
        'ENGINE': 'amazon',
        'INSTANCE_TYPE': 't2.micro'}
    config = femagtools.config.Config(default_config)
    ini_file = os.path.join(os.path.dirname(__file__), 'config.ini')
    config.from_ini_file(ini_file)
    assert config['KEY_NAME'] == 'test-key-pair-eucentral-1'



"""Multiobjective Optimize Algorithm

Nondominated Sorting genetic algorithm II (NSGA-II)

 * NSGA-II is a nondominated-sorting based multiobjective evolutionary algorithm.
 * It genererates offspring with crossover and mutation and select the next
 * generation according to nondominated-sorting and crowding distance comparison.
 *
 * The algorithm can be applied to continuous box-bounded optimization. The version for mixed integer
 * and constrained optimization is also planned.
 *
 * @see Deb, K. and Pratap, A. and Agarwal, S. and Meyarivan, T.,
       "A fast and elitist multiobjective genetic algorithm: NSGA-II"
 """
import random
import numpy as np
from .population import Population
import logging

logger = logging.getLogger("algorithm")


class Nsga2:
    def __init__(self):
        self.cr = 0.95
        self.eta_c = 10
        self.m = 0.01
        self.eta_m = 50
        pass

    def tournament_selection(self, i, j, pop):
        if pop.individuals[i].rank < pop.individuals[j].rank:
            return i
        if pop.individuals[i].rank > pop.individuals[j].rank:
            return j
        if pop.individuals[i].crowd_d < pop.individuals[j].crowd_d:
            return i
        if pop.individuals[i].crowd_d > pop.individuals[j].crowd_d:
            return j
        return random.sample((i, j), 2)[0]

    def crossover(self, pids, pop):
        "return 2 decision vectors"
        logger.debug("crossover {}".format(pids))

        parents = [pop.individuals[i] for i in pids]
        lb = pop.problem.lower
        ub = pop.problem.upper
        child1 = []
        child2 = []
        # a simulated binary crossover SBX
        if np.random.random_sample() <= self.cr:
            for x1, x2, yl, yu in zip(
                    parents[0].cur_x, parents[1].cur_x, lb, ub):
                if np.random.random_sample() <= 0.5 and \
                   abs(x1 - x2) > 1e-14:
                        if x1 < x2:
                            y1 = x1
                            y2 = x2
                        else:
                            y1 = x2
                            y2 = x1

                        rnd = np.random.random_sample()
                        beta = 1.0 + (2.0*(y1-yl)/(y2-y1))
                        alpha = 2.0 - beta**(-(self.eta_c+1.0))
                        if rnd <= (1.0/alpha):
                            betaq = (rnd*alpha)**(1.0/(self.eta_c+1.0))
                        else:
                            betaq = (1.0/(2.0 -
                                          rnd*alpha))**(1.0/(self.eta_c+1.0))

                        c1 = 0.5*((y1+y2)-betaq*(y2-y1))

                        beta = 1.0 + (2.0*(yu-y2)/(y2-y1))
                        alpha = 2.0 - beta**(-(self.eta_c+1.0))
                        if rnd <= (1.0/alpha):
                            betaq = (rnd*alpha)**(1.0/(self.eta_c+1.0))
                        else:
                            betaq = (1.0/(2.0 -
                                          rnd*alpha))**(1.0/(self.eta_c+1.0))

                        c2 = 0.5*((y1+y2)+betaq*(y2-y1))

                        if c1<yl : c1=yl
                        if c2<yl : c2=yl
                        if c1>yu : c1=yu
                        if c2>yu : c2=yu

                        if np.random.random_sample() <= 0.5:
                            child1.append(c1)
                            child2.append(c2)
                        else:
                            child1.append(c2)
                            child2.append(c1)
                else:
                    child1.append(x1)
                    child2.append(x2)
        else:
            child1 = parents[0].cur_x
            child2 = parents[1].cur_x
        return (child1, child2)

    def mutate(self, child, pop):
        mutant = []
        lb = pop.problem.lower
        ub = pop.problem.upper
        for i, x in enumerate(child):
            y = x
            yl = lb[i]
            yu = ub[i]
            delta1 = (y-yl)/(yu-yl)
            delta2 = (yu-y)/(yu-yl)
            rnd = np.random.random_sample()
            mut_pow = 1.0/(self.eta_m+1.0)
            if rnd<=0.5:
                xy = 1.0 - delta1
                val = 2.0*rnd+(1.0-2.0*rnd)*(xy**(self.eta_m+1.0))
                deltaq =  val**mut_pow - 1.0
            else:
                xy = 1.0 - delta2
                val = 2.0*(1.0-rnd)+2.0*(rnd-0.5)*(xy**(self.eta_m+1.0))
                deltaq = 1.0 - val**mut_pow
            y = y + deltaq*(yu-yl)
            if y<yl: y = yl
            if y>yu: y = yu
            mutant.append(y)
        return mutant

    def evolve(self, pop):
        shuffles = (random.sample(range(pop.size()), pop.size()),
                    random.sample(range(pop.size()), pop.size()))
        newpop = Population(pop.problem, 0)
        for i in range(0, pop.size(), 4):
            for s in shuffles:
                parents = (self.tournament_selection(s[i], s[i+1], pop),
                           self.tournament_selection(s[i+2], s[i+3], pop))
                children = self.crossover(parents, pop)
                mutants = (self.mutate(children[0], pop),
                           self.mutate(children[1], pop))
                logger.debug("%s --> %s", children[0], mutants[0])
                logger.debug("%s --> %s", children[1], mutants[1])
                newpop.append(mutants[0])
                newpop.append(mutants[1])
        return newpop



"""
    femagtools.plot.bch
    ~~~~~~~~~~~~~~~~~~~

    Creating bch plots


"""
import numpy as np
import scipy.interpolate as ip
import logging

try:
    import matplotlib
    import matplotlib.pyplot as plt
    import matplotlib.cm as cm
    from mpl_toolkits.mplot3d import Axes3D
    matplotlibversion = matplotlib.__version__
except ImportError:   # ModuleNotFoundError:
    matplotlibversion = 0

logger = logging.getLogger("femagtools.plot.bch")


def _create_3d_axis():
    """creates a subplot with 3d projection if one does not already exist"""
    from matplotlib.projections import get_projection_class
    from matplotlib import _pylab_helpers

    create_axis = True
    if _pylab_helpers.Gcf.get_active() is not None:
        if isinstance(plt.gca(), get_projection_class('3d')):
            create_axis = False
    if create_axis:
        plt.figure()
        plt.subplot(111, projection='3d')


def _plot_surface(ax, x, y, z, labels, azim=None):
    """helper function for surface plots"""
    # ax.tick_params(axis='both', which='major', pad=-3)
    assert np.size(x) > 1 and np.size(y) > 1 and np.size(z) > 1
    if azim is not None:
        ax.azim = azim
    X, Y = np.meshgrid(x, y)
    Z = np.ma.masked_invalid(z)
    ax.plot_surface(X, Y, Z,
                    rstride=1, cstride=1,
                    cmap=cm.viridis, alpha=0.85,
                    vmin=np.nanmin(z), vmax=np.nanmax(z),
                    linewidth=0, antialiased=True)
#                    edgecolor=(0, 0, 0, 0))
    # ax.set_xticks(xticks)
    # ax.set_yticks(yticks)
    # ax.set_zticks(zticks)
    ax.set_xlabel(labels[0])
    ax.set_ylabel(labels[1])
    ax.set_title(labels[2])

    # plt.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0, hspace=0)


def torque(pos, torque, title='', ax=0):
    """creates plot from torque vs position"""
    k = 20
    alpha = np.linspace(pos[0], pos[-1],
                        k*len(torque))
    f = ip.interp1d(pos, torque, kind='quadratic')
    unit = 'Nm'
    scale = 1
    if np.min(torque) < -9.9e3 or np.max(torque) > 9.9e3:
        scale = 1e-3
        unit = 'kNm'
    if ax == 0:
        ax = plt.gca()
    if title:
        ax.set_title(title)
    else:
        ax.set_title('Torque / {}'.format(unit))
    ax.grid(True)
    ax.plot(pos, [scale*t for t in torque], 'go')
    ax.plot(alpha, scale*f(alpha))
    if np.min(torque) > 0 and np.max(torque) > 0:
        ax.set_ylim(bottom=0)
    elif np.min(torque) < 0 and np.max(torque) < 0:
        ax.set_ylim(top=0)


def torque_fft(order, torque, ax=0):
    """plot torque harmonics"""
    unit = 'Nm'
    scale = 1
    if np.min(torque) < -9.9e3 or np.max(torque) > 9.9e3:
        scale = 1e-3
        unit = 'kNm'
    if ax == 0:
        ax = plt.gca()
    ax.set_title('Torque Harmonics / {}'.format(unit))
    ax.grid(True)

    try:
        bw = 2.5E-2*max(order)
        ax.bar(order, [scale*t for t in torque], width=bw, align='center')
        ax.set_xlim(left=-bw/2)
    except ValueError:  # empty sequence
        pass


def force(title, pos, force, xlabel='', ax=0):
    """plot force vs position"""
    unit = 'N'
    scale = 1
    if min(force) < -9.9e3 or max(force) > 9.9e3:
        scale = 1e-3
        unit = 'kN'
    if ax == 0:
        ax = plt.gca()
    ax.set_title('{} / {}'.format(title, unit))
    ax.grid(True)
    ax.plot(pos, [scale*f for f in force])
    if xlabel:
        ax.set_xlabel(xlabel)
    if min(force) > 0:
        ax.set_ylim(bottom=0)


def force_fft(order, force, ax=0):
    """plot force harmonics"""
    unit = 'N'
    scale = 1
    if min(force) < -9.9e3 or max(force) > 9.9e3:
        scale = 1e-3
        unit = 'kN'
    if ax == 0:
        ax = plt.gca()
    ax.set_title('Force Harmonics / {}'.format(unit))
    ax.grid(True)
    try:
        bw = 2.5E-2*max(order)
        ax.bar(order, [scale*t for t in force], width=bw, align='center')
        ax.set_xlim(left=-bw/2)
    except ValueError:  # empty sequence
        pass


def fluxdens_surface(fdens, ax=0):
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    x = [p for p in fdens.positions[0]['X']]
    y = [p['position'] for p in fdens.positions]
    z = np.array([p['B_N']
                  for p in fdens.positions])
    _plot_surface(ax, x, y, z,
                  (u'Rotor pos/°', u'Pos/°', u'B N / T'))


def winding_flux(pos, flux, ax=0):
    """plot flux vs position"""
    if ax == 0:
        ax = plt.gca()
    ax.set_title('Winding Flux / Vs')
    ax.grid(True)
    for p, f in zip(pos, flux):
        ax.plot(p, f)


def winding_current(pos, current, ax=0):
    """plot winding currents"""
    if ax == 0:
        ax = plt.gca()
    ax.set_title('Winding Currents / A')
    ax.grid(True)
    for p, i in zip(pos, current):
        ax.plot(p, i)


def voltage(title, pos, voltage, ax=0):
    """plot voltage vs. position"""
    if ax == 0:
        ax = plt.gca()
    ax.set_title('{} / V'.format(title))
    ax.grid(True)
    ax.plot(pos, voltage)


def voltage_fft(title, order, voltage, ax=0):
    """plot FFT harmonics of voltage"""
    if ax == 0:
        ax = plt.gca()
    ax.set_title('{} / V'.format(title))
    ax.grid(True)
    if max(order) < 5:
        order += [5]
        voltage += [0]
    try:
        bw = 2.5E-2*max(order)
        ax.bar(order, voltage, width=bw, align='center')
    except ValueError:  # empty sequence
        pass


def __get_linearForce_title_keys(lf):
    if 'force_r' in lf:
        return ['Force r', 'Force z'], ['force_r', 'force_z']
    return ['Force x', 'Force y'], ['force_x', 'force_y']


def pmrelsim(bch, title=''):
    """creates a plot of a PM/Rel motor simulation"""
    cols = 2
    rows = 4
    if len(bch.flux['1']) > 1:
        rows += 1
    htitle = 1.5 if title else 0
    fig, ax = plt.subplots(nrows=rows, ncols=cols,
                           figsize=(10, 3*rows + htitle))
    if title:
        fig.suptitle(title, fontsize=16)

    row = 1
    plt.subplot(rows, cols, row)
    if bch.torque:
        torque(bch.torque[-1]['angle'], bch.torque[-1]['torque'])
        plt.subplot(rows, cols, row+1)
        tq = list(bch.torque_fft[-1]['torque'])
        order = list(bch.torque_fft[-1]['order'])
        if order and max(order) < 5:
            order += [15]
            tq += [0]
        torque_fft(order, tq)
        plt.subplot(rows, cols, row+2)
        force('Force Fx',
              bch.torque[-1]['angle'], bch.torque[-1]['force_x'])
        plt.subplot(rows, cols, row+3)
        force('Force Fy',
              bch.torque[-1]['angle'], bch.torque[-1]['force_y'])
        row += 3
    elif bch.linearForce:
        title, keys = __get_linearForce_title_keys(bch.linearForce[-1])
        force(title[0], bch.linearForce[-1]['displ'],
              bch.linearForce[-1][keys[0]], 'Displt. / mm')
        plt.subplot(rows, cols, row+1)
        force_fft(bch.linearForce_fft[-2]['order'],
                  bch.linearForce_fft[-2]['force'])
        plt.subplot(rows, cols, row+2)
        force(title[1], bch.linearForce[-1]['displ'],
              bch.linearForce[-1][keys[1]], 'Displt. / mm')
        plt.subplot(rows, cols, row+3)
        force_fft(bch.linearForce_fft[-1]['order'],
                  bch.linearForce_fft[-1]['force'])
        row += 3

    plt.subplot(rows, cols, row+1)
    flux = [bch.flux[k][-1] for k in bch.flux]
    pos = [f['displ'] for f in flux]
    winding_flux(pos,
                 [f['flux_k'] for f in flux])
    plt.subplot(rows, cols, row+2)
    winding_current(pos,
                    [f['current_k'] for f in flux])
    plt.subplot(rows, cols, row+3)
    voltage('Internal Voltage',
            bch.flux['1'][-1]['displ'],
            bch.flux['1'][-1]['voltage_dpsi'])
    plt.subplot(rows, cols, row+4)
    try:
        voltage_fft('Internal Voltage Harmonics',
                    bch.flux_fft['1'][-1]['order'],
                    bch.flux_fft['1'][-1]['voltage'])
    except:
        pass
    if len(bch.flux['1']) > 1:
        plt.subplot(rows, cols, row+5)
        voltage('No Load Voltage',
                bch.flux['1'][0]['displ'],
                bch.flux['1'][0]['voltage_dpsi'])
        plt.subplot(rows, cols, row+6)
        try:
            voltage_fft('No Load Voltage Harmonics',
                        bch.flux_fft['1'][0]['order'],
                        bch.flux_fft['1'][0]['voltage'])
        except:
            pass
    fig.tight_layout(h_pad=3.5)
    if title:
        fig.subplots_adjust(top=0.92)


def multcal(bch, title=''):
    """creates a plot of a MULT CAL simulation"""
    cols = 2
    rows = 4
    htitle = 1.5 if title else 0
    fig, ax = plt.subplots(nrows=rows, ncols=cols,
                           figsize=(10, 3*rows + htitle))
    if title:
        fig.suptitle(title, fontsize=16)

    row = 1
    plt.subplot(rows, cols, row)
    if bch.torque:
        torque(bch.torque[-1]['angle'], bch.torque[-1]['torque'])
        plt.subplot(rows, cols, row+1)
        tq = list(bch.torque_fft[-1]['torque'])
        order = list(bch.torque_fft[-1]['order'])
        if order and max(order) < 5:
            order += [15]
            tq += [0]
        torque_fft(order, tq)
        plt.subplot(rows, cols, row+2)
        force('Force Fx',
              bch.torque[-1]['angle'], bch.torque[-1]['force_x'])
        plt.subplot(rows, cols, row+3)
        force('Force Fy',
              bch.torque[-1]['angle'], bch.torque[-1]['force_y'])
        row += 3
    elif bch.linearForce:
        title, keys = __get_linearForce_title_keys(bch.linearForce[-1])
        force(title[0], bch.linearForce[-1]['displ'],
              bch.linearForce[-1][keys[0]], 'Displt. / mm')
        plt.subplot(rows, cols, row+1)
        force_fft(bch.linearForce_fft[-2]['order'],
                  bch.linearForce_fft[-2]['force'])
        plt.subplot(rows, cols, row+2)
        force(title[1], bch.linearForce[-1]['displ'],
              bch.linearForce[-1][keys[1]], 'Displt. / mm')
        plt.subplot(rows, cols, row+3)
        force_fft(bch.linearForce_fft[-1]['order'],
                  bch.linearForce_fft[-1]['force'])
        row += 3

    plt.subplot(rows, cols, row+1)
    flux = [bch.flux[k][-1] for k in bch.flux]
    pos = [f['displ'] for f in flux]
    winding_flux(pos,
                 [f['flux_k'] for f in flux])
    plt.subplot(rows, cols, row+2)
    winding_current(pos,
                    [f['current_k'] for f in flux])
    plt.subplot(rows, cols, row+3)
    voltage('Internal Voltage',
            bch.flux['1'][-1]['displ'],
            bch.flux['1'][-1]['voltage_dpsi'])
    plt.subplot(rows, cols, row+4)
    try:
        voltage_fft('Internal Voltage Harmonics',
                    bch.flux_fft['1'][-1]['order'],
                    bch.flux_fft['1'][-1]['voltage'])
    except:
        pass
    if len(bch.flux['1']) > 1:
        plt.subplot(rows, cols, row+5)
        voltage('No Load Voltage',
                bch.flux['1'][0]['displ'],
                bch.flux['1'][0]['voltage_dpsi'])
        plt.subplot(rows, cols, row+6)
        try:
            voltage_fft('No Load Voltage Harmonics',
                        bch.flux_fft['1'][0]['order'],
                        bch.flux_fft['1'][0]['voltage'])
        except:
            pass

    fig.tight_layout(h_pad=3.5)
    if title:
        fig.subplots_adjust(top=0.92)


def fasttorque(bch, title=''):
    """creates a plot of a Fast Torque simulation"""
    cols = 2
    rows = 4
    if len(bch.flux['1']) > 1:
        rows += 1
    htitle = 1.5 if title else 0
    fig, ax = plt.subplots(nrows=rows, ncols=cols,
                           figsize=(10, 3*rows + htitle))
    if title:
        fig.suptitle(title, fontsize=16)

    row = 1
    plt.subplot(rows, cols, row)
    if bch.torque:
        torque(bch.torque[-1]['angle'], bch.torque[-1]['torque'])
        plt.subplot(rows, cols, row+1)
        torque_fft(bch.torque_fft[-1]['order'], bch.torque_fft[-1]['torque'])
        plt.subplot(rows, cols, row+2)
        force('Force Fx',
              bch.torque[-1]['angle'], bch.torque[-1]['force_x'])
        plt.subplot(rows, cols, row+3)
        force('Force Fy',
              bch.torque[-1]['angle'], bch.torque[-1]['force_y'])
        row += 3
    elif bch.linearForce:
        title, keys = __get_linearForce_title_keys(bch.linearForce[-1])
        force(title[0], bch.linearForce[-1]['displ'],
              bch.linearForce[-1][keys[0]], 'Displt. / mm')
        plt.subplot(rows, cols, row+1)
        force_fft(bch.linearForce_fft[-2]['order'],
                  bch.linearForce_fft[-2]['force'])
        plt.subplot(rows, cols, row+2)
        force(title[1], bch.linearForce[-1]['displ'],
              bch.linearForce[-1][keys[1]], 'Displt. / mm')
        plt.subplot(rows, cols, row+3)
        force_fft(bch.linearForce_fft[-1]['order'],
                  bch.linearForce_fft[-1]['force'])
        row += 3

    plt.subplot(rows, cols, row+1)
    flux = [bch.flux[k][-1] for k in bch.flux]
    pos = [f['displ'] for f in flux]
    winding_flux(pos, [f['flux_k'] for f in flux])
    plt.subplot(rows, cols, row+2)
    winding_current(pos, [f['current_k'] for f in flux])
    plt.subplot(rows, cols, row+3)
    voltage('Internal Voltage',
            bch.flux['1'][-1]['displ'],
            bch.flux['1'][-1]['voltage_dpsi'])
    plt.subplot(rows, cols, row+4)
    try:
        voltage_fft('Internal Voltage Harmonics',
                    bch.flux_fft['1'][-1]['order'],
                    bch.flux_fft['1'][-1]['voltage'])
    except:
        pass
    if len(bch.flux['1']) > 1:
        plt.subplot(rows, cols, row+5)
        voltage('No Load Voltage',
                bch.flux['1'][0]['displ'],
                bch.flux['1'][0]['voltage_dpsi'])
        plt.subplot(rows, cols, row+6)
        try:
            voltage_fft('No Load Voltage Harmonics',
                        bch.flux_fft['1'][0]['order'],
                        bch.flux_fft['1'][0]['voltage'])
        except:
            pass
    fig.tight_layout(h_pad=3.5)
    if title:
        fig.subplots_adjust(top=0.92)


def cogging(bch, title=''):
    """creates a cogging plot"""
    cols = 2
    rows = 3

    htitle = 1.5 if title else 0
    fig, ax = plt.subplots(nrows=rows, ncols=cols,
                           figsize=(10, 3*rows + htitle))
    if title:
        fig.suptitle(title, fontsize=16)

    row = 1
    plt.subplot(rows, cols, row)
    if bch.torque:
        torque(bch.torque[0]['angle'], bch.torque[0]['torque'])
        plt.subplot(rows, cols, row+1)
        if bch.torque_fft:
            torque_fft(bch.torque_fft[0]['order'], bch.torque_fft[0]['torque'])
        plt.subplot(rows, cols, row+2)
        force('Force Fx',
              bch.torque[0]['angle'], bch.torque[0]['force_x'])
        plt.subplot(rows, cols, row+3)
        force('Force Fy',
              bch.torque[0]['angle'], bch.torque[0]['force_y'])
        row += 3
    elif bch.linearForce:
        title, keys = __get_linearForce_title_keys(bch.linearForce[-1])
        force(title[0], bch.linearForce[-1]['displ'],
              bch.linearForce[-1][keys[0]], 'Displt. / mm')
        plt.subplot(rows, cols, row+1)
        force_fft(bch.linearForce_fft[-2]['order'],
                  bch.linearForce_fft[-2]['force'])
        plt.subplot(rows, cols, row+2)
        force(title[1], bch.linearForce[-1]['displ'],
              bch.linearForce[-1][keys[1]], 'Displt. / mm')
        plt.subplot(rows, cols, row+3)
        force_fft(bch.linearForce_fft[-1]['order'],
                  bch.linearForce_fft[-1]['force'])
        row += 3

    plt.subplot(rows, cols, row+1)
    voltage('Voltage',
            bch.flux['1'][0]['displ'],
            bch.flux['1'][0]['voltage_dpsi'])
    plt.subplot(rows, cols, row+2)
    voltage_fft('Voltage Harmonics',
                bch.flux_fft['1'][0]['order'],
                bch.flux_fft['1'][0]['voltage'])

    fig.tight_layout(h_pad=2)
    if title:
        fig.subplots_adjust(top=0.92)


def transientsc(bch, title=''):
    """creates a transient short circuit plot"""
    cols = 1
    rows = 2
    htitle = 1.5 if title else 0
    fig, ax = plt.subplots(nrows=rows, ncols=cols,
                           figsize=(10, 3*rows + htitle))
    if title:
        fig.suptitle(title, fontsize=16)

    row = 1
    plt.subplot(rows, cols, row)
    ax = plt.gca()
    ax.set_title('Currents / A')
    ax.grid(True)
    for i in ('ia', 'ib', 'ic'):
        ax.plot(bch.scData['time'], bch.scData[i], label=i)
    ax.set_xlabel('Time / s')
    ax.legend()

    row = 2
    plt.subplot(rows, cols, row)
    ax = plt.gca()
    ax.set_title('Torque / Nm')
    ax.grid(True)
    ax.plot(bch.scData['time'], bch.scData['torque'])
    ax.set_xlabel('Time / s')

    fig.tight_layout(h_pad=2)
    if title:
        fig.subplots_adjust(top=0.92)
    return fig

def transientsc_demag(demag, magnet=0, title='', ax=0):
    """creates a demag plot of a transient short circuit
    Args:
      demag: list of dicts with 'displ', 'H_av', 'H_max', 'lim_hc'
      magnet dict with 'Tmag'
    """
    if ax == 0:
        ax = plt.gca()
    pos = [d['displ'] for d in demag if 'displ' in d]
    hmax = [-d['H_max'] for d in demag if 'H_max' in d]
    havg = [-d['H_av'] for d in demag if 'H_av' in d]
    hclim = [-d['lim_hc'] for d in demag if 'lim_hc' in d]*2

    ax.set_title('Transient Short Circuit Demagnetization [kA/m]')
    ax.plot(pos, hmax,
            label='H Max {:4.2f} kA/m'.format(max(hmax)))
    ax.plot(pos, havg,
            label='H Avg {:4.2f} kA/m'.format(max(havg)))
    ax.plot([pos[0], pos[-1]], hclim, color='C3', linestyle='dashed',
            label='Hc {:4.2f} kA/m'.format(hclim[0]))
    ax.set_xlabel('Rotor Position / °')
    ax.grid(True)
    if magnet:
        ax.legend(title=f"Magnet Temperature {magnet['Tmag']}°C")
    else:
        ax.legend()

def i1beta_torque(i1, beta, torque, title='', ax=0):
    """creates a surface plot of torque vs i1, beta"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    azim = 210
    if 0 < np.mean(beta) or -90 > np.mean(beta):
        azim = -60
    unit = 'Nm'
    scale = 1
    if np.min(torque) < -9.9e3 or np.max(torque) > 9.9e3:
        scale = 1e-3
        unit = 'kNm'
    if title:
        _plot_surface(ax, i1, beta, scale*np.asarray(torque),
                      (u'I1/A', u'Beta/°', title),
                      azim=azim)
    else:
        _plot_surface(ax, i1, beta, scale*np.asarray(torque),
                      (u'I1/A', u'Beta/°', u'Torque/{}'.format(unit)),
                      azim=azim)


def i1beta_ld(i1, beta, ld, ax=0):
    """creates a surface plot of ld vs i1, beta"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    _plot_surface(ax, i1, beta, np.asarray(ld)*1e3,
                  (u'I1/A', u'Beta/°', u'Ld/mH'),
                  azim=60)


def i1beta_lq(i1, beta, lq, ax=0):
    """creates a surface plot of ld vs i1, beta"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    azim = 60
    if 0 < np.mean(beta) or -90 > np.mean(beta):
        azim = -120
    _plot_surface(ax, i1, beta, np.asarray(lq)*1e3,
                  (u'I1/A', u'Beta/°', u'Lq/mH'),
                  azim=azim)


def i1beta_psim(i1, beta, psim, ax=0):
    """creates a surface plot of psim vs i1, beta"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    _plot_surface(ax, i1, beta, psim,
                  (u'I1/A', u'Beta/°', u'Psi m/Vs'),
                  azim=60)


def i1beta_up(i1, beta, up, ax=0):
    """creates a surface plot of up vs i1, beta"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    _plot_surface(ax, i1, beta, up,
                  (u'I1/A', u'Beta/°', u'Up/V'),
                  azim=60)


def i1beta_psid(i1, beta, psid, ax=0):
    """creates a surface plot of psid vs i1, beta"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    azim = -60
    if 0 < np.mean(beta) or -90 > np.mean(beta):
        azim = 60
    _plot_surface(ax, i1, beta, psid,
                  (u'I1/A', u'Beta/°', u'Psi d/Vs'),
                  azim=azim)


def i1beta_psiq(i1, beta, psiq, ax=0):
    """creates a surface plot of psiq vs i1, beta"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    azim = 210
    if 0 < np.mean(beta) or -90 > np.mean(beta):
        azim = -60
    _plot_surface(ax, i1, beta, psiq,
                  (u'I1/A', u'Beta/°', u'Psi q/Vs'),
                  azim=azim)


def idq_torque(id, iq, torque, ax=0):
    """creates a surface plot of torque vs id, iq"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    unit = 'Nm'
    scale = 1
    if np.min(torque) < -9.9e3 or np.max(torque) > 9.9e3:
        scale = 1e-3
        unit = 'kNm'
    _plot_surface(ax, id, iq, scale*np.asarray(torque),
                  (u'Id/A', u'Iq/A', u'Torque/{}'.format(unit)),
                  azim=-60)
    return ax


def idq_psid(id, iq, psid, ax=0):
    """creates a surface plot of psid vs id, iq"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    _plot_surface(ax, id, iq, psid,
                  (u'Id/A', u'Iq/A', u'Psi d/Vs'),
                  azim=210)


def idq_psiq(id, iq, psiq, ax=0):
    """creates a surface plot of psiq vs id, iq"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    _plot_surface(ax, id, iq, psiq,
                  (u'Id/A', u'Iq/A', u'Psi q/Vs'),
                  azim=210)


def idq_psim(id, iq, psim, ax=0):
    """creates a surface plot of psim vs. id, iq"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    _plot_surface(ax, id, iq, psim,
                  (u'Id/A', u'Iq/A', u'Psi m [Vs]'),
                  azim=120)


def idq_ld(id, iq, ld, ax=0):
    """creates a surface plot of ld vs. id, iq"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    _plot_surface(ax, id, iq, np.asarray(ld)*1e3,
                  (u'Id/A', u'Iq/A', u'L d/mH'),
                  azim=120)


def idq_lq(id, iq, lq, ax=0):
    """creates a surface plot of lq vs. id, iq"""
    if ax == 0:
        _create_3d_axis()
        ax = plt.gca()
    _plot_surface(ax, id, iq, np.asarray(lq)*1e3,
                  (u'Id/A', u'Iq/A', u'L q/mH'),
                  azim=120)


def ldlq(bch):
    """creates the surface plots of a BCH reader object
    with a ld-lq identification"""
    beta = bch.ldq['beta']
    i1 = bch.ldq['i1']
    torque = bch.ldq['torque']
    ld = np.array(bch.ldq['ld'])
    lq = np.array(bch.ldq['lq'])
    psid = bch.ldq['psid']
    psiq = bch.ldq['psiq']

    rows = 3
    fig = plt.figure(figsize=(10, 4*rows))
    fig.suptitle('Ld-Lq Identification {}'.format(bch.filename), fontsize=16)
    fig.add_subplot(rows, 2, 1, projection='3d')
    i1beta_torque(i1, beta, torque)

    fig.add_subplot(rows, 2, 2, projection='3d')
    i1beta_psid(i1, beta, psid)

    fig.add_subplot(rows, 2, 3, projection='3d')
    i1beta_psiq(i1, beta, psiq)

    fig.add_subplot(rows, 2, 4, projection='3d')
    try:
        i1beta_psim(i1, beta, bch.ldq['psim'])
    except:
        i1beta_up(i1, beta, bch.ldq['up'])

    fig.add_subplot(rows, 2, 5, projection='3d')
    i1beta_ld(i1, beta, ld)

    fig.add_subplot(rows, 2, 6, projection='3d')
    i1beta_lq(i1, beta, lq)


def psidq(bch):
    """creates the surface plots of a BCH reader object
    with a psid-psiq identification"""
    id = bch.psidq['id']
    iq = bch.psidq['iq']
    torque = bch.psidq['torque']
    ld = np.array(bch.psidq_ldq['ld'])
    lq = np.array(bch.psidq_ldq['lq'])
    psim = bch.psidq_ldq['psim']
    psid = bch.psidq['psid']
    psiq = bch.psidq['psiq']

    rows = 3
    fig = plt.figure(figsize=(10, 4*rows))
    fig.suptitle('Psid-Psiq Identification {}'.format(
        bch.filename), fontsize=16)

    fig.add_subplot(rows, 2, 1, projection='3d')
    idq_torque(id, iq, torque)

    fig.add_subplot(rows, 2, 2, projection='3d')
    idq_psid(id, iq, psid)

    fig.add_subplot(rows, 2, 3, projection='3d')
    idq_psiq(id, iq, psiq)

    fig.add_subplot(rows, 2, 4, projection='3d')
    idq_psim(id, iq, psim)

    fig.add_subplot(rows, 2, 5, projection='3d')
    idq_ld(id, iq, ld)

    fig.add_subplot(rows, 2, 6, projection='3d')
    idq_lq(id, iq, lq)



def main():
    import io
    import sys
    import argparse
    from ..__init__ import __version__
    from femagtools.bch import Reader

    argparser = argparse.ArgumentParser(
        description='Read BCH/BATCH/PLT file and create a plot')
    argparser.add_argument('filename',
                           help='name of BCH/BATCH/PLT file')
    argparser.add_argument(
        "--version",
        "-v",
        action="version",
        version="%(prog)s {}, Python {}".format(__version__, sys.version),
        help="display version information",
    )
    args = argparser.parse_args()
    if not matplotlibversion:
        sys.exit(0)
    if not args.filename:
        sys.exit(0)

    ext = args.filename.split('.')[-1].upper()
    if ext.startswith('MC'):
        import femagtools.mcv
        mcv = femagtools.mcv.read(sys.argv[1])

        if mcv['mc1_type'] in (femagtools.mcv.MAGCRV, femagtools.mcv.ORIENT_CRV):
            ncols = 2
        else:  # Permanent Magnet
            ncols = 1

        fig, ax = plt.subplots(nrows=1, ncols=ncols, figsize=(10, 6))
        if ncols > 1:
            plt.subplot(1, 2, 1)
            mcv_hbj(mcv)
            plt.subplot(1, 2, 2)
            mcv_muer(mcv)
        else:
            mcv_hbj(mcv, log=False)

        fig.tight_layout()
        fig.subplots_adjust(top=0.94)
        plt.show()
        return

    if ext.startswith('PLT'):
        import femagtools.forcedens
        fdens = femagtools.forcedens.read(args.filename)
        cols = 1
        rows = 2
        fig, ax = plt.subplots(nrows=rows, ncols=cols,
                               figsize=(10, 10*rows))
        title = '{}, Rotor position {}'.format(
            fdens.title, fdens.positions[0]['position'])
        pos = fdens.positions[0]['X']
        FT_FN = (fdens.positions[0]['FT'],
                 fdens.positions[0]['FN'])
        plt.subplot(rows, cols, 1)
        forcedens(title, pos, FT_FN)

        title = 'Force Density Harmonics'
        plt.subplot(rows, cols, 2)
        forcedens_fft(title, fdens)

        # fig.tight_layout(h_pad=3.5)
        # if title:
        #    fig.subplots_adjust(top=0.92)
        plt.show()
        return

    bchresults = Reader()
    with io.open(args.filename, encoding='latin1', errors='ignore') as f:
        bchresults.read(f.readlines())

    if (bchresults.type.lower().find(
        'pm-synchronous-motor simulation') >= 0 or
        bchresults.type.lower().find(
            'permanet-magnet-synchronous-motor') >= 0 or
        bchresults.type.lower().find(
            'simulation pm/universal-motor') >= 0):
        pmrelsim(bchresults, bchresults.filename)
    elif bchresults.type.lower().find(
            'multiple calculation of forces and flux') >= 0:
        multcal(bchresults, bchresults.filename)
    elif bchresults.type.lower().find('cogging calculation') >= 0:
        cogging(bchresults, bchresults.filename)
    elif bchresults.type.lower().find('ld-lq-identification') >= 0:
        ldlq(bchresults)
    elif bchresults.type.lower().find('psid-psiq-identification') >= 0:
        psidq(bchresults)
    elif bchresults.type.lower().find('fast_torque calculation') >= 0:
        fasttorque(bchresults)
    elif bchresults.type.lower().find('transient sc') >= 0:
        transientsc(bchresults, bchresults.filename)
    else:
        raise ValueError("BCH type {} not yet supported".format(
            bchresults.type))
    plt.show()


def eigenmode(reigen, num_modes=12):
    """plot eigenmode"""
    cols = 4
    rows = int((num_modes - num_modes%cols)/cols + 1)
    fig, ax = plt.subplots(rows, cols)
    ctr = 0
    for i in range(rows):
        for j in range(cols):
            if ctr < num_modes:
                ax[i, j].imshow(reigen[0][ctr])
                ax[i, j].set_title(f'Freq: {reigen[1][ctr]:.2f} Hz')
                ax[i, j].axis('off')
            else:
                ax[i, j].remove()
                ax[i, j] = None
            ctr+=1
    plt.tight_layout()


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')
    main()



import femagtools.dxfsl.geom as g

import logging


area = [[
    g.Arc(g.Element(center=(93.000391, 104.825783),
                    radius=90.000363,
                    start_angle=-113.674935,
                    end_angle=-102.876410)),
    g.Arc(g.Element(center=(72.676972, 15.919097),
                    radius=1.199853,
                    start_angle=12.363429,
                    end_angle=77.142632)),
    g.Arc(g.Element(center=(0.000000, 0.000000),
                    radius=75.599833,
                    start_angle=10.725943,
                    end_angle=12.355028)),
    g.Arc(g.Element(center=(73.542349, 13.930019),
                    radius=0.749743,
                    start_angle=-99.701978,
                    end_angle=10.759233)),
    g.Arc(g.Element(center=(90.226618, 111.767627),
                    radius=99.999841,
                    start_angle=-110.995630,
                    end_angle=-99.677752)),
    g.Arc(g.Element(center=(75.894867, 74.423271),
                    radius=59.999926,
                    start_angle=-134.554066,
                    end_angle=-110.995752)),
    g.Arc(g.Element(center=(34.501683, 32.380386),
                    radius=0.999352,
                    start_angle=135.002787,
                    end_angle=-134.566333)),
    g.Line(g.Element(start=(36.133000, 35.426000),
                     end=(33.795000, 33.087000))),
    g.Arc(g.Element(center=(36.840535, 34.719239),
                    radius=1.000084,
                    start_angle=45.700702,
                    end_angle=135.031345)),
    g.Arc(g.Element(center=(82.961666, 81.929844),
                    radius=64.999916,
                    start_angle=-134.331700,
                    end_angle=-113.674891))
],
    # deeg example
    [g.Arc(g.Element(center=(2., 6.),
                     radius=1.,
                     start_angle=0.,
                     end_angle=180.)),
     g.Arc(g.Element(center=(6., 2.),
                     radius=1.,
                     start_angle=-90.,
                     end_angle=90.)),
     g.Arc(g.Element(center=(4., 4.),
                     radius=1.,
                     start_angle=180.,
                     end_angle=270.)),
     g.Arc(g.Element(center=(4., 4.),
                     radius=3.,
                     start_angle=180.,
                     end_angle=-90.)),
     g.Line(g.Element(start=(1., 6.), end=(1., 4.0))),
     g.Line(g.Element(start=(3., 4.), end=(3., 6.0))),
     g.Line(g.Element(start=(6., 1.), end=(4., 1.0))),
     g.Line(g.Element(start=(4., 3.), end=(6., 3.0)))]
]


def check_point_inside(n):
    geom = g.Geometry(area[n])
    geom.center = (0.0, 0.0)

    pts = []
    for a in geom.list_of_areas():
        pt = a.get_point_inside(geom)
        assert(pt)
        pts.append(pt)

    assert(pts)
#    p = dr.PlotRenderer()
#    p.render_elements(geom, g.Shape, neighbors=True, points=pts)


def test_0_point_inside():
    check_point_inside(0)


def test_1_point_inside():
    check_point_inside(1)

if __name__ == "__main__":

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')

    test_1_point_inside()



"""
  femagtools.machine.utils

  auxiliary module
"""
import numpy as np
import numpy.linalg as la
import scipy.interpolate as ip
import logging
from .. import parstudy, windings

logger = logging.getLogger(__name__)


def K(d):
    """space phasor transformation matrix
    (Inverse Park Transformation) T-1 * dq
    arguments:
      d: rotation angle

    returns transformation matrix
    """
    return np.array((
        (-np.cos(d), np.sin(d)),
        (-np.cos(d-2*np.pi/3), np.sin(d-2*np.pi/3)),
        (-np.cos(d+2*np.pi/3), np.sin(d+2*np.pi/3))))


def T(d):
    """space phasor transformation matrix
    (Park Transformation) T * abc
    arguments:
      d: rotation angle

    returns transformation matrix
    """
    return np.array((
        (-np.cos(d), -np.cos(d-2*np.pi/3), -np.cos(d+2*np.pi/3)),
        (np.sin(d), np.sin(d-2*np.pi/3), np.sin(d+2*np.pi/3))))/3*2


def invpark(a, q, d):
    """ convert a dq vector to the abc reference frame
    (inverse park transformation)

    Args:
        a: rotation angle
        d: value in direct axis
        q: value in quadrature axis
    """
    if np.isscalar(a) and np.isscalar(q) and np.isscalar(d):
        return np.dot(K(a), (q, d))
    if np.isscalar(q) and np.isscalar(d):
        return np.array([K(x).dot((q, d)) for x in a]).T
    return np.array([K(x).dot((y, z)) for x, y, z in zip(a, d, q)]).T


KTH = 0.0039  # temperature coefficient of resistance
TREF = 20.0  # reference temperature of resistance
EPS = 1e-13


def xiskin(w, temp, zeta, kth=KTH):
    return zeta*np.sqrt(abs(w)/(2*np.pi)/(50*(1+kth*(temp-TREF))))


def kskinl(xi, nl):
    xi2 = 2*xi
    nl2 = nl*nl
    if np.isscalar(xi):
        if xi < EPS:
            k = 1
        else:
            k = (3 / (nl2*xi2)*(np.sinh(xi2) - np.sin(xi2)) /
                 (np.cosh(xi2)-np.cos(xi2)) +
                 ((nl2-1)/(nl2*xi)*(np.sinh(xi)+np.sin(xi)) /
                          (np.cosh(xi)+np.cos(xi))))
    else:
        xi2 = xi2[xi > EPS]
        k = np.ones(np.asarray(xi).shape)
        k[xi > 1e-12] = (3 / (nl2*xi2)*(np.sinh(xi2) - np.sin(xi2)) /
                         (np.cosh(xi2)-np.cos(xi2)) +
                         ((nl2-1)/(nl2*xi)*(np.sinh(xi)+np.sin(xi)) /
                          (np.cosh(xi)+np.cos(xi))))
    return k


def kskinr(xi, nl):
    xi2 = 2*xi
    nl2 = nl*nl
    return xi*((np.sinh(xi2)+np.sin(xi2))/(np.cosh(xi2)-np.cos(xi2))) + \
        ((nl2-1) / 3 * xi2*((np.sinh(xi)-np.sin(xi)) /
                            (np.cosh(xi)+np.cos(xi))))


def wdg_resistance(wdg, n, g, aw, da1, hs, lfe, sigma=56e6):
    """return winding resistance per phase in Ohm
    Arguments:
    wdg: (Winding) winding
    n: (int) number of wires per coil side
    g: (int) number of parallel coil groups
    lfe: length of stator lamination stack in m
    aw: wire cross section area m2
    da1: bore diameter m
    hs: slot height
    sigma: (float) conductivity of wire material 1/Ohm m
    """
    # mean length of one turn
    lt = 2.8*(da1+hs)/2*wdg.yd*2*np.pi/wdg.Q + 16e-3 + 2*lfe
    return wdg.turns_per_phase(n, g)*lt/sigma/aw/g


def wdg_inductance(wdg, n, g, da1, lfe, ag):
    """return winding inductance per phase in H
    Arguments:
    wdg: (Winding) winding
    n: (int) number of wires per coil side
    g: (int) number of parallel coil groups
    da1: bore diameter in m
    lfe: length of stator lamination stack in m
    ag: length of airgap in m
    """
    return wdg.inductance(n, g, da1, lfe, ag)


def skin_resistance(r0, w, temp, zeta, gam=0, nh=1, kth=KTH):
    """return eddy current resistance of winding or rotor bar
    Arguments:
    r0: (float) dc resistance
    w: (float) current frequency in rad
    temp: (float) conductor temperature in deg Celsius
    zeta: (float) skin effect coefficient (penetration depth)
    gam: (float) ratio of length of end winding and length of slot (0..1)
    nh: (int) number of vertical conductors in slot
    kth: (float) temperature coefficient (Default = 0.0039, Cu)"""
    xi = xiskin(w, temp, zeta)
    if np.isscalar(xi):
        if xi < 1e-12:
            k = 1
        else:
            k = (gam + kskinr(xi, nh)) / (1. + gam)
    else:
        k = np.ones(np.asarray(w).shape)
        k[xi > 1e-12] = (gam + kskinr(xi[xi > 1e-12], nh)) / (1. + gam)
    return r0*(1.+kth*(temp - TREF))*k
# return r0*(1.+KTH*(temp - TREF))*(gam + kskinr(xi, nh)) / (1. + gam)


def skin_leakage_inductance(l0, w, temp, zeta, nl=1, pl2v=0.5, kth=KTH):
    """return eddy current leakage inductance of rotor bar
    Arguments:
    l0: (float) dc inductance
    w: (float) current frequency in rad
    temp: (float) conductor temperature in deg Celsius
    zeta: (float) skin effect coefficient (penetration depth)
    nl: (int) number of vertical conductors in slot
    pl2v: (float) variable coefficient (0..1)
    kth: (float) temperature coefficient (Default = 0.0039, Cu)"""
    return l0*(1.0+pl2v*(kskinl(
        xiskin(w, temp, zeta, kth), nl)-1))


def wdg_leakage_inductances(machine):
    """calculate slot leakage and end winding inductances
    ref: Design of Rotating Electrical Machines
    Juha Pyrhönen, Tapani Jokinen, Valeria Hrabovcova
    (Ed. 2008) page 236ff
    """
    from ..windings import Winding
    wdg = Winding(
        {'Q': machine['stator']['num_slots'],
         'm': machine['winding']['num_phases'],
         'p': machine['poles']//2,
         'l': machine['winding']['num_layers'],
         'yd': machine['winding']['coil_span']})
    n1 = wdg.turns_per_phase(machine['winding']['num_wires'],
                             machine['winding']['num_par_wdgs'])
    m = wdg.m
    p = wdg.p
    Q = wdg.Q
    D = machine['bore_diam']
    W = wdg.yd
    taup = Q/2/p
    eps = 1 - W/taup
    slotmodel = [k for k in machine['stator'] if isinstance(
        machine['stator'][k], dict)][-1]
    hs = machine['stator'][slotmodel].get('slot_height',
                                          (machine['outer_diam']-D)/2)
    taus = (D+hs)*np.pi/Q

    b1 = machine['stator'][slotmodel]['slot_width']
    h1 = machine['stator'][slotmodel]['slot_h1']
    h2 = machine['stator'][slotmodel]['slot_h2']
    h3 = 0
    hd = 0
    if machine['stator'][slotmodel].get('tooth_width', 0):
        b4 = taus-machine['stator'][slotmodel]['tooth_width'] + \
            2*machine['stator'][slotmodel].get('slot_r1', 0)
    else:
        b4 = machine['stator'][slotmodel].get('wedge_width1', taus/2) + \
            2*machine['stator'][slotmodel].get('slot_r1', 0)
    h41 = 0
    h42 = h41
    h4 = hs - h1 - h2
    lfe = machine['lfe']
    k1 = 1-9*eps/16
    k2 = 1-3*eps/4
    lbda = k1*(h4-hd)/3/b4 + k2*(h3/b4+h1/b1+h2/(b4-b1)*np.log(b4/b1))+hd/4/b4
    mue0 = 4*np.pi*1e-7
    Lu = 4*m/Q*mue0*lfe*n1**2*lbda
    q = wdg.q
    wew = (1-eps)*(D + h4)*np.pi/2/p
    lew = (lfe-wew)/2
    lmdew = 0.55
    lmw0 = 0.35
    lmdw = 2*lew*lmdew + wew*lmw0
    Lew = 4*m/Q*q*n1**2*mue0*lmdw
    return Lu, Lew


def betai1(iq, id):
    """return beta and amplitude of dq currents"""
    return (np.arctan2(id, iq),
            la.norm((id, iq), axis=0)/np.sqrt(2.0))


def iqd(beta, i1):
    """return qd currents of beta and amplitude"""
    return np.sqrt(2.0)*i1*np.array([np.cos(beta),
                                     np.sin(beta)])


def puconv(dqpar, p, NR, UR, IR):
    """convert dqpar to per unit
    arguments:
    dqpar: dict from ld-iq or psid-psiq identification
    p: pole pairs
    NR: ref speed in 1/s
    UR: ref voltage per phase in V
    IR: ref current per phase in A
    """
    WR = 2*np.pi*NR*p
    PSIR = UR/WR
    SR = 3*UR*IR
    if 'beta' in dqpar:
        dqp = dict(beta=dqpar['beta'], losses=dict())
        dqp['i1'] = np.array(dqpar['i1'])/IR
    elif 'iq' in dqpar:
        dqp = dict(iq=np.array(dqpar['iq)'])/IR*np.sqrt(2), losses=dict())
        dqp['id'] = np.array(dqpar['id'])/IR*np.sqrt(2)
    else:
        raise ValueError('invalid dqpar')
    for k in 'psid', 'psiq':
        dqp[k] = np.array(dqpar[k])/PSIR
    if 'losses' in dqpar:
        for k in ('magnet', 'styoke_hyst', 'styoke_eddy',
                  'stteeth_hyst', 'stteeth_eddy', 'rotor_hyst', 'rotor_eddy'):
            dqp['losses'][k] = np.array(dqpar['losses'][k])/SR
        dqp['losses']['speed'] = p*dqpar['losses']['speed']/WR
        dqp['losses']['ef'] = dqpar['losses']['ef']
        dqp['losses']['hf'] = dqpar['losses']['hf']
    return dqp


def dqpar_interpol(xfit, dqpars, ipkey='temperature'):
    """return interpolated parameters at temperature or exc_current

    Arguments:
      xfit -- temperature or exc_current to fit dqpars
      dqpars -- list of dict with id, iq (or i1, beta), Psid and Psiq values
      ipkey -- key (string) to interpolate
    """
    # check current range
    ckeys = (('i1', 'beta'), ('id', 'iq'))
    dqtype = 0
    fpip = {k: dqpars[0][k] for k in ckeys[dqtype]}
    fpip['losses'] = dict()
    for k in ckeys[dqtype]:
        curr = np.array([f[k] for f in dqpars], dtype=object)
        shape = curr.shape
        if curr.shape != (len(dqpars), len(curr[0])):
            raise ValueError("current range conflict")
        curr = curr.astype(float)
        if not np.array([np.allclose(curr[0], c)
                         for c in curr[1:]]).all():
            raise ValueError("current range conflict")

    try:
        speed = np.array([d['losses']['speed'] for d in dqpars])
        if (np.max(speed) - np.min(speed))/np.mean(speed) > 1e-3:
            raise ValueError("losses: speed conflict")
    except KeyError:
        pass

    sorted_dqpars = sorted(dqpars, key=lambda d: d[ipkey])
    x = [f[ipkey] for f in sorted_dqpars]
    for k in ('psid', 'psiq'):
        m = np.array([f[k] for f in sorted_dqpars]).T
        if len(x) > 2:
            fpip[k] = np.array(
                [[ip.UnivariateSpline(x, y, k=2)(xfit)
                  for y in row] for row in m]).T
        else:
            fpip[k] = ip.interp1d(
                x, m, fill_value='extrapolate')(xfit).T
    try:
        for k in ('styoke_hyst', 'stteeth_hyst',
                  'styoke_eddy', 'stteeth_eddy',
                  'rotor_hyst', 'rotor_eddy',
                  'magnet'):
            m = np.array([f['losses'][k] for f in sorted_dqpars]).T
            if len(x) > 2:
                fpip['losses'][k] = np.array(
                    [[ip.UnivariateSpline(x, y, k=2)(xfit)
                      for y in row] for row in m]).T
            else:
                fpip['losses'][k] = ip.interp1d(
                    x, m, fill_value='extrapolate')(xfit).T
            fpip['losses']['speed'] = dqpars[0]['losses']['speed']
            for f in ('hf', 'ef'):
                if f in dqpars[0]['losses']:
                    fpip['losses'][f] = dqpars[0]['losses'][f]
    except KeyError:
        pass
    return x, fpip


def dqparident(workdir, engine, temp, machine,
               magnetizingCurves, magnetMat=[], condMat=[],
               **kwargs):
    """return list of parameters of equivalent circuit for PM machines

    arguments:
    workdir -- directory for intermediate files
    engine -- calculation driver (multiproc, docker, condor)

    temp -- list of magnet temperatures in degree Celsius
    machine -- dict() with machine parameters
    magnetizingCurves -- list of dict() with BH curves
    magnetMat -- list of dict() with magnet material properties
    condMat -- list of dict() with conductor material properties

    optional arguments:
    num_cur_steps: number of current steps (default 5)
    num_beta_steps: number of current steps (default 7 per quadrant)
    speed: rotor speed in 1/s (default 160/p)
    i1_max: maximum current in A rms (default approx 3*i1nom)
    period_frac: fraction of rotating angle (default 6)
    cmd: femag executable
    """
    import pathlib

    da1 = machine['outer_diam']
    Q1 = machine['stator']['num_slots']
    slotmodel = [k for k in machine['stator'] if isinstance(
        machine['stator'][k], dict)][-1]
    if slotmodel == 'stator1':
        hs = machine['stator']['stator1']['slot_rf1'] - \
            machine['stator']['stator1']['tip_rh1']
    else:
        hs = machine['stator'][slotmodel].get('slot_height', 0)
    wdgk = 'windings' if 'windings' in machine else 'winding'
    N = machine[wdgk]['num_wires']
    Jmax = 15  # max current density in A/mm2

    i1_max = round(0.28*np.pi*hs*(da1+hs)/Q1/N*Jmax*1e5)*10 * \
        machine[wdgk].get('num_par_wdgs', 1)
    period_frac = kwargs.get('period_frac', 6)
    if machine.get('external_rotor', False):
        period_frac = 1  # TODO: missing femag support

    # winding resistance
    wpar = {'Q': machine['stator']['num_slots'],
            'm': machine[wdgk]['num_phases'],
            'p': machine['poles']//2}

    if 'coil_span' in machine[wdgk]:
        wpar['yd'] = machine[wdgk]['coil_span']
    if 'num_layers' in machine[wdgk]:
        wpar['l'] = machine[wdgk]['num_layers']

    wdg = windings.Winding(wpar)

    lfe = machine['lfe']
    g = machine[wdgk].get('num_par_wdgs', 1)
    if 'wire_gauge' in machine[wdgk]:
        aw = machine[wdgk]['wire_gauge']
    elif 'dia_wire' in machine[wdgk]:
        aw = np.pi*machine[wdgk].get('dia_wire', 1e-3)**2/4
    elif ('wire_width' in machine[wdgk]) and ('wire_height' in machine[wdgk]):
        aw = machine['windings']['wire_width']*machine[wdgk]['wire_height']
    else:  # wire diameter from slot area
        aw = 0.75 * machine[wdgk].get('cufilfact', 0.45)*\
            np.pi*da1*hs/Q1/wpar['l']/N
    r1 = wdg_resistance(wdg, N, g, aw, da1, hs, lfe)

    n = len(temp)
    parvardef = {
        "decision_vars": [
            {"values": sorted(2*temp), "name": "magn_temp"},
            {"values": n*[0, -90], "name": "beta_max"},
            {"values": n*[-90, -180], "name": "beta_min"}
        ]
    }

    parvar = parstudy.List(
        workdir, condMat=condMat,
        magnetizingCurves=magnetizingCurves,
        magnets=magnetMat, cmd=kwargs.get('cmd', None))

    leakfile = pathlib.Path(workdir) / 'end_wind_leak.dat'
    leakfile.unlink(missing_ok=True)

    simulation = dict(
        calculationMode='ld_lq_fast',
        i1_max=kwargs.get('i1_max', i1_max),
        magn_temp=20,
        wind_temp=20,
        beta_max=0,
        beta_min=-90,
        num_move_steps=26,
        num_par_wdgs=machine[wdgk].get('num_par_wdgs', 1),
        num_cur_steps=kwargs.get('num_cur_steps', 5),
        num_beta_steps=kwargs.get('num_beta_steps', 7),
        speed=kwargs.get('speed', 160/machine['poles']),
        period_frac=period_frac)

    # TODO: cleanup()  # remove previously created files in workdir
    # start calculation
    results = parvar(parvardef, machine, simulation, engine)
    #import json
    #with open('results.json', 'w') as fp:
    #    json.dump(results, fp)
    ls1 = 0
    try:
        leakages = [float(x)
                    for x in leakfile.read_text().split()]
        ls1 += leakages[1]  # TODO: np.linalg.norm(leakages[1:])
    except:
        logger.warning("No end winding leakage")

    try:
        rotor_mass = sum(results['f'][-1]['weights'][-1])
    except KeyError:
        rotor_mass = 0  # need femag classic > rel-9.3.x-48-gca42bbd0

    ldq = []
    for i in range(0, len(results['f']), 2):
        d = dict(i1=results['f'][i]['ldq']['i1'],
                 beta=results['f'][i+1]['ldq']['beta'][:-1] + results['f'][i]['ldq']['beta'])
        d.update(
            {k: np.vstack((np.array(results['f'][i+1]['ldq'][k])[:-1, :],
                           np.array(results['f'][i]['ldq'][k]))).tolist()
             for k in ('psid', 'psiq', 'torque', 'ld', 'lq', 'psim')})
        ldq.append(d)
    losskeys = ('styoke_hyst', 'stteeth_hyst', 'styoke_eddy',
                'stteeth_eddy', 'rotor_hyst', 'rotor_eddy',
                'magnet')
    for i in range(0, len(results['f']), 2):
        j = i//2
        ldq[j]['temperature'] = results['x'][0][i]
        ldq[j]['losses'] = {k: np.vstack((np.array(results['f'][i+1]['ldq']['losses'][k])[:-1, :],
                                          np.array(results['f'][i]['ldq']['losses'][k]))).tolist()
                            for k in losskeys}
        ldq[j]['losses']['speed'] = results['f'][i]['ldq']['losses']['speed']
        for k in ('hf', 'ef'):
            ldq[j]['losses'][k] = results['f'][i]['lossPar'][k]

    dqpars = {
        'm': machine[wdgk]['num_phases'],
        'p': machine['poles']//2,
        'r1': machine[wdgk].get('resistance', r1),
        'ls1': ls1,
        "rotor_mass": rotor_mass, "kfric_b": 1,
        'ldq': ldq}

    if 'current_angles' in results['f'][0]:
        dqpars['current_angles'] = results['f'][0]['current_angles']
    return dqpars



"""running FEMAG on Amazon cloud EC2

    .. note: To use this engine you have to install the boto3 module from amazon
"""
import os
import threading
import time
import logging

import femagtools.job
from .config import Config

logger = logging.getLogger(__name__)

class MissingConfigurationException(Exception):
    def __init__(self, message):
        Exception.__init__(self, "Missing configuration: {}".format(message))


class Engine(object):

    config_class = Config

    default_config = {
        'ENGINE': 'amazon',
        'SERVER_LOCATION': 'eu-central-1',
        'INSTANCE_TYPE': 't2.micro',
        'ACL': 'authenticated-read',
        'IMAGE_ID': 'ami-b0cc23df',
        'FINISH_TASK_FILENAME': 'exit_code',
        'COMPANY_NAME': 'femag'
    }

    """The Amazon Engine

    This engine uses the boto3 Python module to interact
       with the amazon ec2 and s3 services

    Args:
        buckets (:obj:`list`): Existing buckets with femag calculation files
        configfile (str): Filename of config file

    .. :note: If possible you should use the same location for all services

    """
    def __init__(self, buckets=None, configfile='config.ini'):
        self.buckets = buckets
        self.job = None

        # Amazon file storage
        self.s3_resource = self._create_amazon_resource('s3')
        # Amazon Server administration
        self.ec2_resource = self._create_amazon_resource('ec2')

        # Create instance of config
        self.config = Config(self.default_config)
        self.config.from_ini_file(configfile)

    def _create_amazon_resource(self, resource):
        import boto3
        return boto3.resource(resource)

    def _create_data_buckets(self):
        """Create unique S3 Buckets for calculation

        Args:
            ACL (str): ACL-Rules for Amazon
        """
        # If buckets exsists map them with a folder
        if self.buckets:
            for idx, bucket in enumerate(self.buckets):
                self.job.tasks[idx].id = bucket['id']
                self.job.tasks[idx].directory = bucket['folder']
            return

        bucketConfiguration = {'LocationConstraint': self.config['SERVER_LOCATION']}

        # Create a bucket for every calculation
        for t in self.job.tasks:
            self.s3_resource.create_bucket(ACL=self.config['ACL'],
                                           Bucket=t.id,
                                           CreateBucketConfiguration=bucketConfiguration)

        logger.debug("Created buckets")

    def _upload_files_to_s3(self):
        """Upload all files to Amazon S3 for this calculation

        """
        if self.buckets:
            logger.info("Files are already uploaded")
            return

        threads = []
        for t in self.job.tasks:
            thread = threading.Thread(target=self._upload, args=(t, ))
            threads.append(thread)
            thread.start()

        logger.info("Uploading files: ")
        self._wait_for_threads_finished(threads, "Upload files")

    def _upload(self, task):
        """Upload thread for uploading one directory

        :internal:

        Args:
            task (py:class:`CloudTask`): The task which belongs to the uploading folder
        """
        # Upload one single tar_file
        task.tar_file.close()

        name = os.path.basename(task.file)
        Body = open(task.file, 'rb')
        self.s3_resource.Object(task.id, name).put(Body=Body)

    def _wait_for_threads_finished(self, threads, operation):
        """Wait until all threads are finished

        :internal:

        Args:
            threads (:obj:`list`): List of threads to check if they are finished
            operation (str): Name of the operation to write a meaningful log message

        """
        # Wait until all threads are not running
        while not all([t.isAlive() is False for t in threads]):
            time.sleep(5)

        # timer.cancel
        logger.info("{} is finished".format(operation))

    def _start_instances(self):
        """Start all instances for the calculation

        """
        # Prepare arguemtns for instance start
        param = {'MinCount': 1, 'MaxCount': 1 }

        if self.config.get('IMAGE_ID', None):
            param['ImageId'] = self.config['IMAGE_ID']
        else:
            raise MissingConfigurationException('image_id')

        if self.config.get('INSTANCE_TYPE', None):
            param['InstanceType'] = self.config['INSTANCE_TYPE']
        else:
            raise MissingConfigurationException('instance_type')

        if self.config.get('IAM_INSTANCE_PROFILE', None):
            param['IamInstanceProfile'] = {'Name': self.config['IAM_INSTANCE_PROFILE'] }

        if self.config.get('KEY_NAME', None):
            param['KeyName'] = self.config['KEY_NAME']

        # Set security group id as list
        if self.config['SECURITY_GROUP_IDS']:
            param['SecurityGroupIds'] = []
            for security_group in [s for s in self.config['SECURITY_GROUP_IDS'] if s]:
                param['SecurityGroupIds'].append(security_group)

        param['DryRun'] = self.config.get('DRY_RUN', False)

        threads = []
        for idx, t in enumerate(self.job.tasks):
            thread = threading.Thread(target=self._start_instance, args=(param, t))
            threads.append(thread)
            thread.start()

        self._wait_for_threads_finished(threads, "Start instances")

    def _start_instance(self, param, task):
        """Start one instance

        :internal:

        Args:
            task (Task): the task for calculation

        """
        user_data = self._read_cloud_init(task.id)
        if user_data:
            param['UserData'] = user_data

        instances = self.ec2_resource.create_instances(**param)

        instance = instances[0]  # We only started one instance
        logger.info("Instance started: {}".format(instance.id))
        self._add_tag(task.id, instance.id)

        instance.wait_until_running()
        instance.load()  # Reload the data to get public dns etc.
        logger.info("Instance {} is running: Public dns: {}".format(instance.id, instance.public_dns_name))
        task.ec2_instance = instance.id

    def _add_tag(self, task_id, instance_id):
        """Add a tag to the instance

        :internal:

        Args:
            task_id (int): The task id (Same as the S3 Bucket name)
            instance_id (int): The instance_id to set the tag to the right instance
        """
        tag = '{}-{}'.format(task_id, self.config.get('COMPANY_NAME', 'femag'))
        self.ec2_resource.create_tags(Resources=[instance_id], Tags=[{'Key': 'Name', 'Value': tag}])

    def _read_cloud_init(self, bucket_name):
        """Read the cloud init file and if there is a line which starts with {{ENV}}
        then put all config options as environment variables.

        """
        user_data = ""
        # Set all config options as environment variable
        if os.path.isfile(self.config.get('CLOUD_INIT', None)):
            with open(self.config['CLOUD_INIT'], 'rt') as f:
                for line in f:
                    if line.startswith('{{ENV}}'):
                        # Add config
                        for key, value in sorted(self.config.items()):
                            user_data += "export {}={}\n".format(key, value)
                        # add other important stuff
                        user_data += "export BUCKET_NAME={}\n".format(bucket_name)
                        continue
                    user_data += line
        return user_data

    def _join(self, timeout=20, filename='exit_code'):
        """Wait until all instances are finished with the calulation.

        :internal:

        Args:
            timeout (int): How long we wait between a check
            filename (str): What is the filename of the exit_code
        """
        import botocore       # For exception

        finished_tasks = []
        client = self.s3_resource.meta.client
        while len(finished_tasks) < len(self.job.tasks):
            for t in [task for task in self.job.tasks if task not in finished_tasks]:
                try:
                    client.get_object(Bucket=t.id, Key=filename)
                except botocore.exceptions.ClientError:
                    # Instance not ready
                    time.sleep(2)
                    continue
                finished_tasks.append(t)
                logger.info("Calculation is finished for instance {}".format(t.id))
                self.ec2_resource.instances.filter(InstanceIds=[t.ec2_instance]).terminate()
            time.sleep(timeout)

        logger.info("Calculations are finished")

    def _get_result_data_from_S3(self):
        """Get all the calculated files to the correct folder

        """
        import boto3
        client = self.s3_resource.meta.client
        transfer = boto3.s3.transfer.S3Transfer(client)
        for t in self.job.tasks:
            bucket = t.id
            folder = t.directory
            files = client.list_objects(Bucket=bucket)['Contents']
            logger.debug("Starting new folder")

            for file in files:
                file_name = file['Key']
                transfer.download_file(bucket, file_name, os.path.join("{}/{}".format(folder, file_name)))
                logger.debug("Downloaded file {}".format(file_name))

    def _get_status_code(self, filename='exit_code'):
        """Get the status code from the caluclation

        Args:
            filename (str): Filename of exit_code
        """
        status_code = []
        for t in self.job.tasks:
            dir = "{}/{}".format(t.directory, filename)
            file = open(dir, 'r')
            status_code.append(file.read())
        return status_code

    def _cleanup(self):
        threads = []
        for t in self.job.tasks:
            thread = threading.Thread(target=self._delete_bucket, args=(t.id, ))
            threads.append(thread)
            thread.start()

        logger.info("Deleting buckets: ")
        self._wait_for_threads_finished(threads, "Deleting buckets")

        # Clean up volumes
        client = self.ec2_resource.meta.client
        volumes = client.describe_volumes(Filters=[{'Name': 'status', 'Values': ['available']}])['Volumes']
        for v in volumes:
            client.delete_volume(VolumeId=v['VolumeId'])

    def _delete_bucket(self, bucket_name):
        bucket = self.s3_resource.Bucket(bucket_name)
        for key in bucket.objects.all():
            key.delete()
        bucket.delete()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # FEMAG STUFF
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    def create_job(self, workdir):
        """Create a FEMAG :py:class:`CloudJob`

        Args:
            workdir (str): The workdir where the calculation files are stored

        Return:
            Cloud job (:class:`CloudJob`)
        """
        self.job = femagtools.job.CloudJob(workdir)
        return self.job

    def submit(self):
        """Starts the FEMAG calculation(s) on Amazon

        Return:
            length of started tasks (int)
        """
        self._create_data_buckets()
        self._upload_files_to_s3()
        self._start_instances()
        return len(self.job.tasks)

    def join( self ):
        """Wait until all calculations are finished

        Return:
            list of all calculations status (C = Ok, X = error) (:obj:`list`)
        """
        status = []
        # Wait until all tasks are finished
        self._join(timeout=20, filename=self.config['FINISH_TASK_FILENAME'])

        # get all files
        self._get_result_data_from_S3()

        # Remove buckets if cleanup is set
        if int(self.config.get('DELETE_BUCKETS', 0)):
            self._cleanup()

        status = self._get_status_code(filename=self.config['FINISH_TASK_FILENAME'])
        for t, r in zip(self.job.tasks, status):
            t.status = 'C' if int(r)==0 else 'X'

        return status



import femagtools.machine.im
import pathlib
import pytest
import numpy as np


@pytest.fixture
def im():
    impars = {
        "p": 2, "m": 3, "f1type": 50, "u1type": 230, "rotor_mass": 6, "kfric_b": 1,
        "r1": 0.2, "r2": 0.5389, "lsigma1": 2e-3,
        "lsigma2": 4.4e-3, "kh": 10, "zeta2": 1,
        "fec": 120, "fee": 0, "fexp": 7.0,
        "iml": 5.317456519935536, "ims": 3.425169199057511, "mexp": 8.75033417091787}
    return femagtools.machine.im.InductionMachine(impars)


def test_im(im):
    u1 = 230
    f1 = 50
    s = 0.01
    torque = im.torqueu(2*np.pi*f1, u1, 2*np.pi*(1-s)*f1/im.p)
    assert pytest.approx(torque, rel=0.01) == 17.77



#!/usr/bin/env python
#
import femagtools.job
import tempfile
import os


def test_condor():
    workdir = tempfile.mkdtemp()
    job = femagtools.job.CondorJob(workdir)
    task = job.add_task()
    task.add_file('femag.fsl', ['exit_on_end=True'])
    try:
        job.prepareDescription()
    
        with open(os.path.join(workdir, 'femag.submit')) as f:
            x = [l.strip().split('=') for l in f]

        len(x) == 12
        x[-1][0].split() == '1'

        d = {l[0].strip(): l[1].strip() for l in x if len(l) > 1}
        d['Universe'] == 'vanilla'
    except Exception as e:
        pass # no femag installed on ci server.
    

def test_job():
    workdir = tempfile.mkdtemp()
    job = femagtools.job.Job(workdir)
    task = job.add_task()
    task.add_file('femag.fsl', ['exit_on_end=True'])
    
    with open(os.path.join(workdir, '0/femag.fsl')) as f:
        x = [l.strip().split('=') for l in f]
    d = {k: v for k, v in x}
    d['exit_on_end'] == 'True'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
import femagtools.conductor

condMat = [dict(
    name="Cu",
    spmaweight=8.96,
    elconduct=56e6,
    tempcoef=3.9e-3)
]


def test_findById():
    cond = femagtools.conductor.Conductor(condMat)
    result = cond.find('Cu')
    expected = condMat[0]
    assert result == expected



# -*- coding: utf-8 -*-
""" Configuration of FEMAG interface

"""
import platform
import os
try:
    # Load python 2 config parser
    import ConfigParser as configparser
except ImportError:
    import configparser

executable = {
    'Linux': {
        'mag_static': 'xfemag64',
        'mag_dynamic': 'cxfemag64',
        'mag_transient': 'tsfemag64',
        'mech_static': 'mefemag64',
        'therm_static': 'thfemag64'},
    'Windows': {
        'mag_static': 'wfemagw64.exe',
        'mag_dynamic': 'wcfemagw64.exe',
        'mag_transient': 'wtsfemagw64.exe',
        'mech_static': 'wmefemagw64.exe',
        'therm_static': 'wthfemagw64.exe'}
}

state_of_problem_set = set(executable['Linux'].keys())
"""set of state of problem keys"""

def is_exe(fpath):
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)


def which(program):
    """returns complete pathname of program or None if not found"""
    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None


def get_executable(stateofproblem='mag_static'):
    """returns platform specific pathname of femag executable"""
    progname = executable[platform.system()][stateofproblem]
    femag = which(progname)
    if not femag:
        raise ValueError(f"{progname} not found in {os.environ['PATH']}")
        # TODO: for l in ['libstdc++-6.dll', 'libgcc_s_dw2-1.dll']:
        #    lib = which(l)
        #    if not lib:
        #        raise Exception("{} not found in {}".format(
        #            l, os.environ["PATH"]))

        #    transfer_files.append(lib)
    return femag


class Config(dict):

    def __init__(self, defaults=None, configfile=None):
        dict.__init__(self, defaults or {})

    def from_ini_file(self, ini_file):
        if not ini_file:
            return
        config = configparser.ConfigParser()
        config.read(ini_file)

        # Only get config for this engine
        section = self['ENGINE']

        try:
            if config.items(section.lower()):
                engine = section.lower()
            else:
                if config.items(section.upper()):
                    engine = section.upper()
                else:
                    return
        except configparser.NoSectionError:
            try:
                config[section.upper()]
                engine = section.upper()
            except:
                return

        for key, value in config.items(engine):
            if ',' in value:
                self[key.upper()] = value.split(',')
                continue
            self[key.upper()] = value

    def from_ini_file_all(self, ini_file):
        if not ini_file:
            return

        config = configparser.ConfigParser()
        config.read(ini_file)

        for group in config:
            for key in config[group]:
                self[key.upper()] = config[group][key]



import femagtools.dxfsl.geom as g
import numpy as np
import logging
import pytest


def test_alpha_points():
    p1 = (3.0, 1.0)
    p2 = (4.0, 5.0)
    p3 = (1.0, 3.0)

    alfa1 = g.alpha_points(p1, p2, p3)
    alfa2 = g.alpha_points(p2, p3, p1)
    alfa3 = g.alpha_points(p3, p1, p2)

    alfa = alfa1 + alfa2 + alfa3
    assert(np.isclose(alfa, 2*np.pi))


def test_points_are_close():
    rtol = 1e-05
    atol = 1e-08
    p1 = (3.14, 2.7666)
    p2 = (3.141592, 2.77)

    A = g.points_are_close(p1, p2, rtol, atol)
    B = g.points_are_close(p1, p2, rtol, 0.001)
    assert(A == B)

    A = g.points_are_close(p1, p2, 1e-01, atol)
    B = g.points_are_close(p1, p2, 1e-02, atol)
    assert(A == B)


def test_same_angle():
    assert(g.is_same_angle(3.14159266, np.pi))
    assert(g.is_same_angle(3.14159266, -np.pi))


def test_misc():
    assert(g.gcd(4, 5) == 1)
    assert(g.gcd(24, 6) == 6)
    assert(g.gcd(47, 23) == 1)
    assert(g.gcd(18, 3) == 3)
    assert(g.gcd(39, 13) == 13)


if __name__ == "__main__":

    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s %(message)s')

    test_alpha_points()
    test_points_are_close()
    test_same_angle()
    test_misc()



"""post processing based on vtu-files created by FEMAG-TS
"""
__author__ = 'werner b. vetter, ronald tanner'

import femagtools.nc
import femagtools.vtu as vtu
import numpy as np
import scipy.integrate as integrate
import warnings
import pathlib
import logging

logger = logging.getLogger('femagtools.asm')


def _read_sections(lines):
    """return list of TS sections

    sections may be surrounded by lines starting with 'Windings/branches'
    Args:
      param lines (list) lines of TS file to read

    Returns:
      list of sections
    """

    section = []
    title = ''
    group = []
    columns = []
    if lines:
        title = lines[0]
        s = 1
        if 'Groups' in lines[1]:
            s = 2
        for line in lines[s:]:
            i = 0
            if 'Windings/branches' in line:
                i = 1
                yield section[i:]
                section = []
            else:
                section.append(line.strip())
    yield section


def read_st(arg, modelname):
    """read TS result files stv, stc, stm
    Arguments:
    arg: (str or Path) name of directory
    modelname: (str) name of model
    """
    r = {}
    if isinstance(arg, str):
        dir = pathlib.Path(arg)
    elif isinstance(arg, pathlib.Path):
        dir = arg
    else:
        logger.error("Invalid arg type %s", type(arg))
        return
    res = dict()
    for ext in ('stv', 'stc', 'stm'):
        lines = (dir / f"{modelname}.{ext}").read_text().split('\n')
        for s in _read_sections(lines):
            if s:
                k = 2
                if ext == 'stm':
                    k = 1
                    labels = [l.lower()
                              for l in s[k-1].split()[::2]][1:]
                elif ext == 'stc':
                    labels = [f'I{l}'
                              for l in s[k-1].split()[1::2]][1:]
                else:
                    labels = [f'U{l}'
                              for l in s[k-1].split()[1::2]][1:]
                m = np.array(
                    [[float(x) for x in l.split()]
                     for l in s[k:] if l]).T
                if 'time' not in res:
                    res['time'] = m[0].tolist()
                for i, k in enumerate(labels):
                    res[k] = m[i+1].tolist()
    return res


def losscoeff_frequency_to_time(B0, f0, c, exp):
    '''Convert Bertotti-coefficient of frequency domain to time domains coefficient
        Parameters
        ----------
        B0 : float
            Base flux density [T]
        f0 : float
            Base freuency [Hz]
        c : float
            Bertotti-coefficient
        exp : float
            Bertotti-exponent
        Return
        -------
        k : float
            Loss coefficient in time domains

        The conversion is only possible for loss-coefficients with
        equal exponent for flux density and frequency,
        as eddy current losses (cw*(B/B0(*)*2*(f/f0)**2) or
        anomalous losses (ce*(B/B0(*)**1.5*(f/f0)**1.5)
        '''
    y, abserr = integrate.quad(lambda x: np.abs(np.cos(2*np.pi*x))**exp, 0, 1)
    return c/(B0**exp*f0**exp)/((2*np.pi)**exp * y)


class TimeRange(object):
    def __init__(self, vtu_data, nc_model):
        '''Read time vector in and generate an equidistant vector if necessary.
           Also the base frequency is determined.
        Parameters
        ----------
        vtu_data : object
            vtu reader
        nc_model: object
        '''

        try:  # FEMAG-TS files
            data_list = ['time [s]']
            vtu_data.read_data(data_list)
            self.vector = vtu_data.get_data_vector('time [s]')
            self.freq = 1/(self.vector[-1]-self.vector[0] +
                           (self.vector[1]-self.vector[0])/2 +
                           (self.vector[-1]-self.vector[-2])/2)
            dt = self.vector[1]-self.vector[0]
            dt_min = 1e32
            self.equidistant = True
            for i in range(len(self.vector)-2):
                dti = self.vector[i+1]-self.vector[i]
                if dt < 0.999*dti or dt > 1.001*dti:
                    self.equidistant = False
                if dti < dt_min:
                    dt_min = dti
            if not self.equidistant:
                numpnt = int((self.vector[-1]-self.vector[0])/dt_min)
                self.vector_equi = np.linspace(self.vector[0],
                                               self.vector[-1],
                                               num=numpnt)
        except:  # FEMAG-DC files
            speed = nc_model.speed
            self.freq = speed/60*nc_model.pole_pairs
            self.equidistant = True


class Losses(object):
    def __init__(self, modelname, dirname):
        '''Loss calculation for FEMAG-TS simulations
        Parameters
        ----------
        dirname : str
            Name of the model (nc-file)
        ncmodel : object

        '''
        self.vtu_data = vtu.read(dirname)
        self.nc_model = femagtools.nc.read(modelname)
        # Read iron losses coefficients
        self.iron_loss_coefficients = self.nc_model.iron_loss_coefficients
        for c in self.iron_loss_coefficients:
            if c['cw_freq_exp'] == c['cw_ind_exp']:
                c['kw'] = losscoeff_frequency_to_time(
                    c['base_induction'],
                    c['base_frequency'],
                    c['cw'], c['cw_freq_exp'])
            else:
                warnings.warn(
                    'Waterfall method not possible, specify parameter kw')
                kw = 0.0

    def ohm_lossenergy_el(self, el, supel):
        '''Ohmic loss energy of an element
        Parameters
        ----------
        el: object
            Element
        supel: object
            Superelement of the element (for material data)

        Returns
        -------
        ellossenergy : float
            Ohmic loss energy of the element
        '''

        length = self.nc_model.arm_length
        time = self.times.vector

        ff = supel.fillfactor
        if ff == 0.0:
            ff = 1.0

        temp_corr = 1+supel.temp_coef*(el.temperature-20)
        ellossenergy = 0.0
        cd_vec = self.vtu_data.get_data_vector('curd', el.key)
        for j in range(len(time)-1):
            cd = (cd_vec[j]+cd_vec[j+1])/2
            dt = time[j+1]-time[j]
            ellossenergy = ellossenergy + dt*cd**2*el.area/ff / \
                supel.conduc*temp_corr*supel.length
        return ellossenergy*length

    def ohm_lossenergy_sr(self, sr):
        '''Ohmic loss energy of a subregion
        Parameters
        ----------
        sr : object
            Subregion

        Returns
        -------
        lossenergy : float
            Ohmic loss energy of the subregion

        The loss energy is determined by adding up the loss energy of the
        individual elements.
        '''
        scale_factor = self.nc_model.scale_factor()
        srlossenergy = 0.0
        for se in sr.superelements:
            selossenergy = 0.0
            if se.conduc > 0.0:
                for el in se.elements:
                    ellossenergy = self.ohm_lossenergy_el(el, se)
                    selossenergy = selossenergy + ellossenergy * scale_factor

            srlossenergy = srlossenergy + selossenergy

        return srlossenergy

    def ohm_lossenergy_subregion(self, srname, start=0.0, end=0.0):
        '''Ohmic loss energy of a subregion
        Parameters
        ----------
        srname:  str
            Name of subregion
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)

        Returns
        -------
        lossenergy : float
            Ohmic loss energy of the subregion

        The loss energy is determined by adding up the loss energy of the
        individual elements over the time window.
        If start and end are not specified, the time window of the
        previous calculation is used.
        '''
        data_list = ['time [s]', 'curd']
        self.vtu_data.read_data(data_list)

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start, end)

        self.times = TimeRange(self.vtu_data, self.nc_model)

        sr = self.nc_model.get_subregion(srname)
        return self.ohm_lossenergy_sr(sr)

    def ohm_powerlosses_subregion(self, srname, start=0.0, end=0.0):
        '''Ohmic loss dissipation of a subregion within the time window
        Parameters
        ----------
        srname : str
            Name of subregion
        start : float
            Start of the time window (optional)
        end : float
            End of the time window (optional)

        Returns
        -------
        powerlosses : float
            Ohmic loss dissipation of the subregion

        The loss energy is determined by adding up the loss energy of the
        individual elements over the time window.
        The loss energy is divided by the time window length
        to obtain the averaged power loss
        If start and end are not specified, the time window of the
        previous calculation is used.
        '''
        while len(srname) < 4:
            srname = srname+' '

        time = self.times.vector[-1]-self.times.vector[0]
        return srlossenergy / time

    def ohm_lossenergy(self, start=0.0, end=0.0):
        '''Ohmic loss energy of all subregions
        Parameters
        ----------
        start: float
            Start of the time window (optional)
        end: float
            End of the time window (optional)

        Returns
        -------
        loss_data: dict
            Dictonary of subregions and ohmic loss energy of it

        The loss energy is determined by adding up the loss energy of the
        individual elements over the time window.
        If start and end are not specified, the time window of the
        previous calculation is used.
        '''

        data_list = ['time [s]', 'curd']
        self.vtu_data.read_data(data_list)

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start, end)

        self.times = TimeRange(self.vtu_data, self.nc_model)

        loss_data = []
        for sr in self.nc_model.subregions:
            srlossenergy = self.ohm_lossenergy_sr(sr)

            srname = sr.name
            if sr.wb_key >= 0:
                if srname == '    ':
                    srname = "wdg "+str(sr.wb_key+1)

            loss_data.append(
                {'key': sr.key, 'name': srname, 'losses': srlossenergy})

        return loss_data

    def ohm_powerlosses(self, start=0.0, end=0.0):
        '''Ohmic loss dissipation of all subregions
        Parameters
        ----------
        start : float
            Start of the time window (optional)
        end : float
            End of the time window (optional)

        Returns
        -------
        loss_data : dict
            Dictonary of subregions and ohmic loss dissipation of it

        The loss energy is determined by adding up the loss energy of the
        individual elements over the time window.
        The loss energy is divided by the time window length
        to obtain the averaged power loss
        If start and end are not specified, the time window of the
        previous calculation is used.
        '''

        data_list = ['time [s]', 'curd']
        self.vtu_data.read_data(data_list)

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start, end)

        self.times = TimeRange(self.vtu_data, self.nc_model)
        #self.time_vector = self.vtu_data.get_data_vector('time [s]')
        time = self.times.vector[-1]-self.times.vector[0]

        loss_data = []
        for sr in self.nc_model.subregions:
            srlossenergy = self.ohm_lossenergy_sr(sr)
            srpowerlosses = srlossenergy / time

            srname = sr.name
            if sr.wb_key >= 0:
                if srname == '    ':
                    srname = "wdg "+str(sr.wb_key+1)

            loss_data.append(
                {'key': sr.key, 'name': srname, 'losses': srpowerlosses})

        return loss_data

    def ohm_powerlosses_fft_el(self, el, supel):
        '''Power dissipation of an element
        Parameters
        ----------
        el : object
            Element
        supel: object
            Superelement of the element (for material data)

        Returns
        -------
        elpowerlosses : float
            Ohmic power losses of the element

        A FFT from the current density is made.
        The power losses of each harmonic is determined and added.
        '''

        length = self.nc_model.arm_length

        ff = supel.fillfactor
        if ff == 0.0:
            ff = 1.0

        temp_corr = 1+supel.temp_coef*(el.temperature-20)
        elpowerlosses = 0.0
        cd_vec_0 = self.vtu_data.get_data_vector('curd', el.key)
        if not self.times.equidistant:
            cd_vec = np.interp(self.times.vector_equi,
                               self.times.vector, cd_vec_0,
                               period=1.0/self.times.freq)
        else:
            cd_vec = cd_vec_0
        cd_spec = abs(np.fft.fft(cd_vec))/(len(cd_vec)/2)
        for j in range(int(len(cd_vec)/2)):
            elpowerlosses = elpowerlosses + \
                cd_spec[j]**2/2*el.area/ff / \
                supel.conduc*temp_corr*supel.length

        return elpowerlosses*length

    def ohm_powerlosses_fft_sr(self, sr):
        '''Power dissipation of a subregion
        Parameters
        ----------
        sr : object
            Subregion

        Returns
        -------
        powerlosses : float
            Ohmic power losses of the subregion

        A FFT from the current density is made.
        The power losses of each harmonic is determined and added.
        '''
        scale_factor = self.nc_model.scale_factor()

        srpowerlosses = 0.0
        for se in sr.superelements:
            sepowerlosses = 0.0
            if se.conduc > 0.0:
                for el in se.elements:
                    elpowerlosses = self.ohm_powerlosses_fft_el(el, se)
                    sepowerlosses = sepowerlosses + elpowerlosses * scale_factor

            srpowerlosses = srpowerlosses + sepowerlosses

        return srpowerlosses

    def ohm_powerlosses_fft_subregion(self, srname, start=0.0, end=0.0):
        '''Power dissipation of a subregion
        Parameters
        ----------
        srname:  str
            Name of subregion
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)

        Returns
        -------
        lossenergy : float
            Power dissipation of the subregion

        A FFT from the current density is made.
        The power losses of each harmonic is determined and added.
        The time window has to be pariode or a multiple of it.
        If start and end are not specified, the time window of the
        previous calculation is used.
        '''
        data_list = ['time [s]', 'curd']
        self.vtu_data.read_data(data_list)

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start, end)

        self.times = TimeRange(self.vtu_data, self.nc_model)

        sr = self.nc_model.get_subregion(srname)
        srpowerlosses = self.ohm_powerlosses_fft_sr(sr)
        return srpowerlosses

    def ohm_powerlosses_fft(self, start=0.0, end=0.0):
        '''Power dissipation of all subregions
        Parameters
        ----------
        start : float
            Start of the time window (optional)
        end : float
            End of the time window (optional)

        Returns
        -------
        loss_data : dict
            Dictonary of subregions and power dissipation of it

        A FFT from the current density is made.
        The power losses of each harmonic is determined and added.
        The time window has to be pariode or a multiple of it.
        If start and end are not specified, the time window of the
        previous calculation is used.
        '''

        data_list = ['time [s]', 'curd']
        self.vtu_data.read_data(data_list)

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start, end)

        self.times = TimeRange(self.vtu_data, self.nc_model)

        loss_data = []
        for sr in self.nc_model.subregions:
            srpowerlosses = self.ohm_powerlosses_fft_sr(sr)

            srname = sr.name
            if sr.wb_key >= 0:
                if srname == '    ':
                    srname = "wdg "+str(sr.wb_key+1)

            loss_data.append(
                {'key': sr.key, 'name': srname, 'losses': srpowerlosses})

        return loss_data

# iron losses
    def iron_losses_fft_el(self, el, se):
        '''Iron losses of an element
        Parameters
        ----------
        el: object
            Element
        se: object
            Superelement of element (for material data)

        Returns
        -------
        ironlosses : float
            Iron losses of the element

        A FFT is made from the flux density.
        The iron losses of each harmonic is determined  by
        Bertotti formula

            Physt = ch * (f/f0)**hfe * (B/B0)**hBe * V * rho * shape_factor
            Peddy = ch * (f/f0)**wfe * (B/B0)**wBe * V * rho * shape_factor
            Pexce = ch * (f/f0)**efe * (B/B0)**eBe * V * rho * shape_factor

        and added to the total losses of the element

            Ptot  = (Physt + Peddy + Pexce)
        '''

        length = self.nc_model.arm_length
        freq = self.times.freq

        eltotallosses = 0.0
        elhystlosses = 0.0
        eleddylosses = 0.0
        elexcelosses = 0.0

        if (se.elements[0].reluc[0] < 1.0 or se.elements[0].reluc[1] < 1.0) and \
           (se.elements[0].mag[0] == 0.0 and se.elements[0].mag[1] == 0.0):

            center_pnt = se.elements[0].center
            try:
                if (np.sqrt(center_pnt[0]**2+center_pnt[1]**2) > self.nc_model.FC_RADIUS):
                    ldi = len(self.iron_loss_coefficients)-2  # outside
                else:
                    ldi = len(self.iron_loss_coefficients)-1  # inside
            except:
                ldi = len(self.iron_loss_coefficients)-1  # inside
            #sf = self.iron_loss_coefficients[ldi]['shapefactor']
            if (se.mcvtype > 0):
                ldi = se.mcvtype-1
            bf = self.iron_loss_coefficients[ldi]['base_frequency']
            bb = self.iron_loss_coefficients[ldi]['base_induction']
            ch = self.iron_loss_coefficients[ldi]['ch']
            chfe = self.iron_loss_coefficients[ldi]['ch_freq_exp']
            chbe = self.iron_loss_coefficients[ldi]['ch_ind_exp']
            cw = self.iron_loss_coefficients[ldi]['cw']
            cwfe = self.iron_loss_coefficients[ldi]['cw_freq_exp']
            cwbe = self.iron_loss_coefficients[ldi]['cw_ind_exp']
            ce = self.iron_loss_coefficients[ldi]['ce']
            cefe = self.iron_loss_coefficients[ldi]['ce_freq_exp']
            cebe = self.iron_loss_coefficients[ldi]['ce_ind_exp']
            sw = self.iron_loss_coefficients[ldi]['spec_weight']*1000
            ff = self.iron_loss_coefficients[ldi]['fillfactor']

            bx_vec_0 = self.vtu_data.get_data_vector('b', el.key)[0]
            if not self.times.equidistant:
                bx_vec = np.interp(self.times.vector_equi,
                                   self.times.vector, bx_vec_0,
                                   period=1.0/self.times.freq)
                # f = interpolate.interp1d(self.times.vector, bx_vec_0, kind="cubic")
                # bx_vec = f(self.times.vector_equi)
            else:
                bx_vec = bx_vec_0
            bx_vec = np.array(bx_vec)/ff
            spx = np.fft.fft(bx_vec)
            bx_spec = abs(spx)/(len(bx_vec)/2)
            bx_spec[0] = bx_spec[0]/2
            bx_phi = np.arctan2(spx.imag, spx.real)

            by_vec_0 = self.vtu_data.get_data_vector('b', el.key)[1]
            if not self.times.equidistant:
                by_vec = np.interp(self.times.vector_equi,
                                   self.times.vector, by_vec_0,
                                   period=1.0/self.times.freq)
                # f = interpolate.interp1d(self.times.vector, by_vec_0, kind="cubic")
                # by_vec = f(self.times.vector_equi)
            else:
                by_vec = by_vec_0
            by_vec = np.array(by_vec)/ff
            spy = np.fft.fft(by_vec)
            by_spec = abs(spy)/(len(by_vec)/2)
            by_spec[0] = by_spec[0]/2
            by_phi = np.arctan2(spy.imag, spy.real)

            b_abs = np.sqrt(np.array(bx_vec)**2+np.array(by_vec)**2)
            b_spec = np.sqrt(bx_spec**2+by_spec**2)

            # kh: Korrekturfaktor für drehendes Feld
            kz = 1.0
            i_max = list(b_spec).index(max(b_spec))
            # Phasenverschiebung zwischen Grundschwingungskomponeneten
            dphi = bx_phi[i_max]-by_phi[i_max]
            while dphi > np.pi/2:
                dphi = dphi - np.pi
            while dphi < -np.pi/2:
                dphi = dphi + np.pi
            # Gleichanteil
            b_dc = max(abs(bx_spec[0]), abs(by_spec[0]))
            # Achsenverhältnis der Ellipse
            axis = 0.0
            #Transformation in Hauptrichtung
            j_max = list(b_abs).index(max(b_abs))
            phi = np.arctan2(by_vec[j_max], bx_vec[j_max])
            bxt_vec = []
            byt_vec = []
            for i in range(len(bx_vec)):
                bxt_vec.append(np.cos(phi)*bx_vec[i]+np.sin(phi)*by_vec[i])
                byt_vec.append(np.sin(phi)*bx_vec[i]-np.cos(phi)*by_vec[i])
            # Achsenverhältnis
            max_bxt = max(abs(np.array(bxt_vec)))
            max_byt = max(abs(np.array(byt_vec)))
            if (max_byt > 1.0e-3):
                axis = max_bxt/max_byt
            if axis > 1.0:
                axis = 1.0/axis

            # Korrekturfaktor
            if ((abs(dphi) > np.pi/3) and axis > 0.3):
                kz = 1.55
            if (b_dc > 0.2):
                kz = 1.0 + 0.65*b_dc**2.1
            if (max(b_spec) > 1.85):
                kz = 1.1

            for j in range(int(len(b_spec)/2)):
                elhystlosses = elhystlosses + kz * ch * \
                    (j*freq/bf)**chfe*(b_spec[j]/bb)**chbe
                eleddylosses = eleddylosses + cw * \
                    (j*freq/bf)**cwfe*(b_spec[j]/bb)**cwbe
                elexcelosses = elexcelosses + ce * \
                    (j*freq/bf)**cefe*(b_spec[j]/bb)**cebe

            elhystlosses = elhystlosses*el.area*length*ff*sw
            eleddylosses = eleddylosses*el.area*length*ff*sw
            elexcelosses = elexcelosses*el.area*length*ff*sw

        eltotallosses = elhystlosses + eleddylosses + elexcelosses

        return {'total': eltotallosses,
                'hysteresis': elhystlosses,
                'eddycurrent': eleddylosses,
                'excess': elexcelosses}

    def iron_losses_fft_se(self, se):
        '''Iron losses of a superelement
        Parameters
        ----------
        se: object
            Superelement

        Returns
        -------
        ironlosses : float
            Iron losses of the superelement

        The iron losses are calculated based on the Bertotti formula
        (see also iron_losses_fft_el)
        The results are muliplied by the scale factor,
        represent also the whole machine.
        '''

        scale_factor = self.nc_model.scale_factor()

        sehystlosses = 0.0
        seeddylosses = 0.0
        seexcelosses = 0.0
        if (se.elements[0].reluc[0] < 1.0 or se.elements[0].reluc[1] < 1.0) and \
                (se.elements[0].mag[0] == 0.0 and se.elements[0].mag[1] == 0.0):
            for el in se.elements:
                ellosses = self.iron_losses_fft_el(el, se)
                sehystlosses = sehystlosses + \
                    ellosses['hysteresis'] * scale_factor
                seeddylosses = seeddylosses + \
                    ellosses['eddycurrent'] * scale_factor
                seexcelosses = seexcelosses + ellosses['excess'] * scale_factor

        setotallosses = sehystlosses + seeddylosses + seexcelosses

        return {'total': setotallosses,
                'hysteresis': sehystlosses,
                'eddycurrent': seeddylosses,
                'excess': seexcelosses}

    def iron_losses_fft_subregion(self, srname, start=0.0, end=0.0):
        '''Iron losses of a subregion
        Parameters
        ----------
        srname:  str
            Name of subregion
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)

        Returns
        -------
        losses : dict
            Iron losses of the subregion

        The iron losses are calculated based on the Bertotti formula
        (see also ron_losses_fft_se)
        '''
        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start, end)

        data_list = ['b']
        self.vtu_data.read_data(data_list)
        self.times = TimeRange(self.vtu_data, self.nc_model)

        srtotallosses = 0.0
        srhystlosses = 0.0
        sreddylosses = 0.0
        srexcelosses = 0.0
        sr = self.nc_model.get_subregion(srname)
        for se in sr.superelements:
            selosses = self.iron_losses_fft_se(se)
            srtotallosses = srtotallosses + selosses['total']
            srhystlosses = srhystlosses + selosses['hysteresis']
            sreddylosses = sreddylosses + selosses['eddycurrent']
            srexcelosses = srexcelosses + selosses['excess']

        srlosses = {'subregion': srname,
                    'total': srtotallosses,
                    'hysteresis': srhystlosses,
                    'eddycurrent': sreddylosses,
                    'excess': srexcelosses
                    }
        return srlosses

    def iron_losses_fft(self, start=0.0, end=0.0):
        '''Iron losses of all subregion and superelements
        Parameters
        ----------
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)

        Returns
        -------
        losses : dict
            Iron losses of the subregion

        The iron losses are calculated based on the Bertotti formula
        (see also iron_losses_fft_se)
        '''

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start, end)

        data_list = ['b']
        self.vtu_data.read_data(data_list)
        self.times = TimeRange(self.vtu_data, self.nc_model)

        losseslist = []

        for se in self.nc_model.superelements:
            selosses = self.iron_losses_fft_se(se)

            if se.subregion:
                for sr in self.nc_model.subregions:
                    if se in sr.superelements:
                        srname = sr.name
            else:
                if (se.mcvtype == 0):
                    center_pnt = se.elements[0].center
                    if (np.sqrt(center_pnt[0]**2+center_pnt[1]**2) > self.nc_model.FC_RADIUS):
                        srname = "no, outside"
                    else:
                        srname = "no, inside"

            found = False
            for srlosses in losseslist:
                if srlosses['subregion'] == srname:
                    srlosses['total'] = srlosses['total']+selosses['total']
                    srlosses['hysteresis'] = srlosses['hysteresis'] + \
                        selosses['hysteresis']
                    srlosses['eddycurrent'] = srlosses['eddycurrent'] + \
                        selosses['eddycurrent']
                    srlosses['excess'] = srlosses['excess']+selosses['excess']
                    found = True
            if not found:
                if selosses['total'] > 0.0:
                    srlosses = {'subregion': srname,
                                'total': selosses['total'],
                                'hysteresis': selosses['hysteresis'],
                                'eddycurrent': selosses['eddycurrent'],
                                'excess': selosses['excess']
                                }
                    losseslist.append(srlosses)

        return losseslist

    def iron_lossenergy_time_el(self, el, se):
        '''Iron losses of an elemt in time domain
        Parameters
        ----------
        el: object
             Element
        se: object
             Superelement of element

        Returns
        -------
        lossenergies : float
             Iron losses of the element

        The iron losses are calculated based on the Bertotti formula
        in time domaine.
        The loss coefficients in frequency domain are converted into
        time domain coefficients.
        For the hysteresis losses is a water fall methode implemented.
        Eddy current losses and anomalous losses are calculated by
        add up the losses of each time step.
        '''

        length = self.nc_model.arm_length
        time = self.times.vector

        elhystenergy = 0.0
        eleddyenergy = 0.0
        elexceenergy = 0.0

        if (se.elements[0].reluc[0] < 1.0 or se.elements[0].reluc[1] < 1.0) and \
                (se.elements[0].mag[0] == 0.0 and se.elements[0].mag[1] == 0.0):
            if (se.mcvtype == 0):
                center_pnt = se.elements[0].center
                try:
                    if (np.sqrt(center_pnt[0]**2+center_pnt[1]**2) > self.nc_model.FC_RADIUS):
                        ldi = len(self.iron_loss_coefficients)-2  # outside
                    else:
                        ldi = len(self.iron_loss_coefficients)-1  # inside
                except:
                    ldi = len(self.iron_loss_coefficients)-1  # use inside
            else:
                ldi = se.mcvtype-1
            kh = self.iron_loss_coefficients[ldi]['kh']
            chbe = self.iron_loss_coefficients[ldi]['ch_ind_exp']
            khml = self.iron_loss_coefficients[ldi]['khml']
            kw = self.iron_loss_coefficients[ldi]['kw']
            cwbe = self.iron_loss_coefficients[ldi]['cw_ind_exp']
            ke = self.iron_loss_coefficients[ldi]['ke']
            cebe = self.iron_loss_coefficients[ldi]['ce_ind_exp']
            sw = self.iron_loss_coefficients[ldi]['spec_weight']*1000
            ff = self.iron_loss_coefficients[ldi]['fillfactor']
            #sf = self.iron_loss_coefficients[ldi]['shapefactor']

            bx_vec = np.array(self.vtu_data.get_data_vector('b', el.key)[0])/ff
            by_vec = np.array(self.vtu_data.get_data_vector('b', el.key)[1])/ff

            # Maximalwert und Richtung des Haupfeldes
            Bpeak = np.sqrt(bx_vec[0]**2+by_vec[0]**2)
            phi = np.arctan2(by_vec[0], bx_vec[0])
            for i in range(1, len(time)):
                b1 = np.sqrt(bx_vec[i-1]**2+by_vec[i-1]**2)
                b2 = np.sqrt(bx_vec[i]**2+by_vec[i]**2)
                if abs(b2) > Bpeak:
                    Bpeak = abs(b2)
                    phi = np.arctan2(by_vec[i], bx_vec[i])

            #Transformation in Hauptrichutng
            br_vec = []
            bt_vec = []
            for i in range(len(time)):
                br_vec.append(np.cos(phi)*bx_vec[i]+np.sin(phi)*by_vec[i])
                bt_vec.append(np.sin(phi)*bx_vec[i]-np.cos(phi)*by_vec[i])

            Bpeak_p = np.sqrt(bx_vec[0]**2+by_vec[0]**2)
            Bx = []
            tp_beg = 0.0
            tp_end = 0.0
            Tp = 0.0
            nzeros = 0
            zero = (br_vec[0] >= 0)
            if br_vec[1] > br_vec[0]:
                up = True
            else:
                up = False
            for i in range(1, len(time)):
                b1 = np.sqrt(br_vec[i-1]**2+bt_vec[i-1]**2)
                b2 = np.sqrt(br_vec[i]**2+bt_vec[i]**2)
                # Maximalwert innerhalb letzter Periode
                if abs(b2) > Bpeak_p:
                    Bpeak_p = abs(b2)
                # Nulldurchgaenge und Periodendauer
                if zero != (br_vec[i] >= 0):
                    zero = (not zero)
                    tp_beg = tp_end
                    tp_end = time[i]
                    if tp_beg > 0.0:
                        nzeros = nzeros+1
                        if nzeros > 1:
                            #Tp = (Tp*(nzeros-1)/nzeros+2*(tp_end-tp_beg)/nzeros)/2
                            Tp = 2*(tp_end-tp_beg)
                            Bpeak = Bpeak_p
                            elhystenergy = elhystenergy+kh*Bpeak**chbe/2
                            Bpeak_p = 0.0
                        else:
                            Tp = 2.0*(tp_end-tp_beg)
                            Bpeak = Bpeak_p
                            elhystenergy = elhystenergy+kh * \
                                Bpeak**chbe * (tp_end-time[0])/Tp
                            Bpeak_p = 0.0
                    Bx = []
                # Wendepunkte
                if up and b2 < b1:
                    Bx.append(b1)
                if not up and b2 > b1:
                    Bx.append(b1)
                # Steigungsrichtung
                if b2 > b1:
                    up = True
                else:
                    up = False

                try:
                    if b2 > 0 and up and b2 > Bx[-2]:
                        Bm = abs(Bx[-2]+Bx[-1])/2
                        dB = abs(Bx[-2]-Bx[-1])
                        elhystenergy = elhystenergy + \
                            kh*Bm**(chbe-1)*khml*dB/2
                        Bx.remove(Bx[-2])
                        Bx.remove(Bx[-1])
                    if b2 < 0 and not up and b2 < Bx[-2]:
                        elhystenergy = elhystenergy + \
                            kh*Bm**(chbe-1)*khml*dB/2
                        Bx.remove(Bx[-2])
                        Bx.remove(Bx[-1])
                    if b2 > 0 and not up and Bx[-1] > Bx[-2]:
                        elhystenergy = elhystenergy + \
                            kh*Bm**(chbe-1)*khml*dB/2
                        Bx.remove(Bx[-2])
                        Bx.remove(Bx[-1])
                    if b2 < 0 and up and Bx[-1] < Bx[-2]:
                        elhystenergy = elhystenergy + \
                            kh*Bm**(chbe-1)*khml*dB/2
                        Bx.remove(Bx[-2])
                        Bx.remove(Bx[-1])

                except:
                    pass

                dt = time[i]-time[i-1]
                dbr = br_vec[i]-br_vec[i-1]
                dbt = bt_vec[i]-bt_vec[i-1]
                db = np.sqrt(dbr**2+dbt**2)
                eleddyenergy = eleddyenergy + kw*(db/dt)**cwbe * dt
                elexceenergy = elexceenergy + ke*(db/dt)**cebe * dt

            #elhystenergy = elhystenergy+kh*Bpeak**chbe * T/(time[-1]-time[0])
            if nzeros >= 1:
                elhystenergy = elhystenergy+kh * \
                    Bpeak**chbe * (time[-1]-tp_end)/Tp

            elhystenergy = elhystenergy*el.area*length*ff*sw
            eleddyenergy = eleddyenergy*el.area*length*ff*sw
            elexceenergy = elexceenergy*el.area*length*ff*sw

        eltotalenergy = elhystenergy + eleddyenergy + elexceenergy

        return {'total': eltotalenergy,
                'hysteresis': elhystenergy,
                'eddycurrent': eleddyenergy,
                'excess': elexceenergy}

    def iron_lossenergy_time_se(self, se):
        '''Iron losses of a superelement in time domain
        Parameters
        ----------
        se: object
            Superelement

        Returns
        -------
        lossenergies : float
            Iron losses of the superelement

        The iron losses are calculated based on the Bertotti formula
        in time domain (see also iron_lossenergy_time_el)
        The results are muliplied by the scale factor,
        represent also the whole machine.
        '''

        scale_factor = self.nc_model.scale_factor()

        sehystenergy = 0.0
        seeddyenergy = 0.0
        seexceenergy = 0.0
        if (se.elements[0].reluc[0] < 1.0 or se.elements[0].reluc[1] < 1.0) and \
                (se.elements[0].mag[0] == 0.0 and se.elements[0].mag[1] == 0.0):
            for el in se.elements:
                elenergy = self.iron_lossenergy_time_el(el, se)
                sehystenergy = sehystenergy + \
                    elenergy['hysteresis'] * scale_factor
                seeddyenergy = seeddyenergy + \
                    elenergy['eddycurrent'] * scale_factor
                seexceenergy = seexceenergy + elenergy['excess'] * scale_factor

        setotalenergy = sehystenergy + seeddyenergy + seexceenergy

        return {'total': setotalenergy,
                'hysteresis': sehystenergy,
                'eddycurrent': seeddyenergy,
                'excess': seexceenergy}

    def iron_lossenergy_time_subregion(self, srname, start=0.0, end=0.0):
        '''Iron loss energy of a subregion
        Parameters
        ----------
        srname:  str
            Name of subregion
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)

        Returns
        -------
        losses : dict
            Iron losses energy of the subregion

        The iron losses are calculated based on the Bertotti formula
        in time domain (see also iron_lossenergy_time_el)
        '''
        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start, end)

        data_list = ['b']
        self.vtu_data.read_data(data_list)
        self.times = TimeRange(self.vtu_data, self.nc_model)

        srtotalenergy = 0.0
        srhystenergy = 0.0
        sreddyenergy = 0.0
        srexceenergy = 0.0
        sr = self.nc_model.get_subregion(srname)
        for se in sr.superelements:
            seenergy = self.iron_lossenergy_time_se(se)
            srtotalenergy = srtotalenergy + seenergy['total']
            srhystenergy = srhystenergy + seenergy['hysteresis']
            sreddyenergy = sreddyenergy + seenergy['eddycurrent']
            srexceenergy = srexceenergy + seenergy['excess']

        return {'subregion': srname,
                'total': srtotalenergy,
                'hysteresis': srhystenergy,
                'eddycurrent': sreddyenergy,
                'excess': srexceenergy
                }

    def iron_losses_time_subregion(self, srname, start=0.0, end=0.0):
        '''Iron power losses of a subregion
        Parameters
        ----------
        srname:  str
            Name of subregion
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)

        Returns
        -------
        losses : dict
            Iron losses energy of the subregion

        The iron losses are calculated based on the Bertotti formula
        in time domain (see also iron_lossenergy_time_el)
        '''
        while len(srname) < 4:
            srname = srname+' '

        srenergy = self.iron_lossenergy_time_subregion(srname, start, end)
        time = self.times.vector[-1]-self.times.vector[0]

        srlosses = {'subregion': srname,
                    'total': srenergy['total']/time,
                    'hysteresis': srenergy['hysteresis']/time,
                    'eddycurrent': srenergy['eddycurrent']/time,
                    'excess': srenergy['excess']/time
                    }

        return srlosses

    def iron_lossenergy_time(self, start=0.0, end=0.0):
        '''Iron losses of all subregion and superelements
        Parameters
        ----------
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)

        Returns
        -------
        energies : dict
            Iron losses enegies of the subregion

        The iron losses are calculated based on the Bertotti formula
        in time domain (see also iron_lossenergy_time_se)
        '''

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start, end)

        data_list = ['b']
        self.vtu_data.read_data(data_list)
        self.times = TimeRange(self.vtu_data, self.nc_model)

        energylist = []

        for se in self.nc_model.superelements:
            selossenergy = self.iron_lossenergy_time_se(se)

            if se.subregion:
                for sr in self.nc_model.subregions:
                    if se in sr.superelements:
                        srname = sr.name
            else:
                if (se.mcvtype == 0):
                    center_pnt = se.elements[0].center
                    try:
                        if (np.sqrt(center_pnt[0]**2+center_pnt[1]**2) > self.nc_model.FC_RADIUS):
                            srname = "no, outside"
                        else:
                            srname = "no, inside"
                    except:
                        srname = "no, used inside"

            found = False
            for srlosses in energylist:
                if srlosses['subregion'] == srname:
                    srlosses['total'] = srlosses['total']+selossenergy['total']
                    srlosses['hysteresis'] = srlosses['hysteresis'] + \
                        selossenergy['hysteresis']
                    srlosses['eddycurrent'] = srlosses['eddycurrent'] + \
                        selossenergy['eddycurrent']
                    srlosses['excess'] = srlosses['excess'] + \
                        selossenergy['excess']
                    found = True
            if not found:
                if selossenergy['total'] > 0.0:
                    srlosses = {'subregion': srname,
                                'total': selossenergy['total'],
                                'hysteresis': selossenergy['hysteresis'],
                                'eddycurrent': selossenergy['eddycurrent'],
                                'excess': selossenergy['excess']
                                }
                    energylist.append(srlosses)

        return energylist

    def iron_losses_time(self, start=0.0, end=0.0):
        '''Iron losses of all subregion and superelements
        Parameters
        ----------
        start: float
            Start of the time window (optional)
        end : float
            End of the time window (optional)

        Returns
        -------
        losses : dict
            Iron losses of the subregion

        The iron losses are calculated based on the Bertotti formula
        in time domain (see also iron_lossenergy_time_el)
        '''

        energylist = self.iron_lossenergy_time(start, end)
        time = self.times.vector[-1]-self.times.vector[0]

        losseslist = []
        for sr in energylist:
            sr['total'] = sr['total']/time
            sr['hysteresis'] = sr['hysteresis']/time
            sr['eddycurrent'] = sr['eddycurrent']/time
            sr['excess'] = sr['excess']/time
            losseslist.append(sr)

        return losseslist

    def export_lossdensity(self,  filename, methode="fft", start=0.0, end=0.0):
        '''Export the loss density of elements in a vtu -file

        Args:
          filename: string Filename of created vtu-file (with extension)
        methode: string Calculation methode (optional, default="fft")
            methode="fft": use fft to calculate the losses
            methode="time": calculate the losses in time domain
        start: float  Start of the time window (optional)
        end : float  End of the time window (optional)

        Returns:
          The losses density in each element is calculated and stored in a vtu-file.
        '''

        import vtk

        if start != 0.0 or end != 0.0:
            self.vtu_data.set_time_window(start, end)

        #scale_factor = self.nc_model.scale_factor()
        length = self.nc_model.arm_length
        self.times = TimeRange(self.vtu_data, self.nc_model)
        time = self.times.vector[-1]-self.times.vector[0]

        dest_grid = vtk.vtkUnstructuredGrid()
        # copy points
        num_point = self.vtu_data.output.GetNumberOfPoints()
        points = vtk.vtkPoints()
        for i in range(num_point):
            pnt = self.vtu_data.output.GetPoints().GetPoint(i)
            points.InsertNextPoint(pnt)
        dest_grid.SetPoints(points)
        # copy cells
        num_cells = self.vtu_data.output.GetNumberOfCells()
        cells = vtk.vtkCellArray()
        cell_types = []
        # original cells

        for i in range(num_cells):
            cell = self.vtu_data.output.GetCell(i)
            if cell.GetNumberOfPoints() == 3:
                triangle = vtk.vtkTriangle()
                for j in range(cell.GetNumberOfPoints()):
                    triangle.GetPointIds().SetId(j, cell.GetPointId(j))
                cells.InsertNextCell(triangle)
                cell_types.append(vtk.VTK_TRIANGLE)
            if cell.GetNumberOfPoints() == 4:
                quad = vtk.vtkQuad()
                for j in range(cell.GetNumberOfPoints()):
                    quad.GetPointIds().SetId(j, cell.GetPointId(j))
                cells.InsertNextCell(quad)
                cell_types.append(vtk.VTK_QUAD)
        dest_grid.SetCells(cell_types, cells)

        # insert cell values
        cell_data = vtk.vtkDoubleArray()
        cell_data.SetNumberOfComponents(1)
        cell_data.SetName("lossdensity [W/m3]")
        cell_data.SetNumberOfValues(num_cells)

        for se in self.nc_model.superelements:
            for el in se.elements:
                ellossdensity = 0
                if se.conduc > 0.0:
                    if methode == "time":
                        ellossenergy = self.ohm_lossenergy_el(el, se)
                        ellossdensity = ellossenergy / \
                            (time * el.area * length)
                    else:
                        ellosses = self.ohm_powerlosses_fft_el(el, se)
                        ellossdensity = ellosses / (el.area * length)

                if (se.elements[0].reluc[0] < 1.0 or se.elements[0].reluc[1] < 1.0) and \
                   (se.elements[0].mag[0] == 0.0 and se.elements[0].mag[1] == 0.0):
                    if methode == "time":
                        ellossenergy = self.iron_lossenergy_time_el(el, se)
                        ellossdensity = ellossdensity + \
                            ellossenergy['total'] / (time * el.area * length)
                    else:
                        ellosses = self.iron_losses_fft_el(el, se)
                        ellossdensity = ellossdensity + \
                            ellosses['total'] / (el.area * length)

                cell_data.InsertValue(el.key-1, ellossdensity)

        dest_grid.GetCellData().AddArray(cell_data)

        writer = vtk.vtkXMLUnstructuredGridWriter()
        writer.SetFileName(filename)
        writer.SetInputData(dest_grid)
        writer.Update()
        writer.Write()

        return



""" Axial Flux PM Machine

"""
import logging
import numpy as np
from pathlib import Path
from .. import poc
from .. import parstudy
from .. import model
from .. import utils
from .. import windings
from scipy.interpolate import RegularGridInterpolator, interp1d
from scipy.integrate import quad

logger = logging.getLogger(__name__)

AFM_TYPES = (
    "S1R1",      # 1 stator, 1 rotor
    "S1R2",      # 1 stator, 2 rotor, 1 half simulated
    "S1R2_all",  # 1 stator, 2 rotor, all simulated
    "S2R1",      # 2 stator, 1 rotor, 1 half simulated
    "S2R1_all"   # 2 stator, 1 rotor, all simulated
)

def _integrate(radius, pos, val):
    interp = RegularGridInterpolator((radius, pos), val)
    def func(x, y):
        return interp((x, y))
    return [quad(func, radius[0], radius[-1], args=(p,))[0]
            for p in pos]


def _integrate1d(radius, val):
    interp = interp1d(radius, val)
    def func(x):
        return interp((x))
    return quad(func, radius[0], radius[-1])[0]


def parident(workdir, engine, temp, machine,
             magnetizingCurves, magnetMat=[], condMat=[],
             **kwargs):
    """return list of parameters of equivalent circuit for AFPM

    arguments:
    workdir: (str) directory for intermediate files
    engine: (object) calculation driver (multiproc, docker, condor)

    temp: list of magnet temperatures in degree Celsius
    machine: dict() with machine parameters
    magnetizingCurves: list of dict() with BH curves
    magnetMat: list of dict() with magnet material properties
    condMat: list of dict() with conductor material properties

    optional arguments:
    num_slices: number of slices (default 3)
    num_cur_steps: number of current steps (default 5)
    num_beta_steps: number of current steps (default 7 per quadrant)
    speed: rotor speed in 1/s (default 160/p)
    i1_max: maximum current in A rms (default approx 3*i1nom)
    cmd: femag executable
    """
    try:
        if machine['afmtype'] not in AFM_TYPES:
            raise ValueError(f"invalid afm type {machine['afmtype']}")
    except KeyError:
        raise ValueError("missing key afmtype")

    di = machine['inner_diam']
    Q1 = machine['stator']['num_slots']
    slotmodel = 'afm_stator'
    hs = machine['stator'][slotmodel]['slot_height']
    wdgk = 'windings' if 'windings' in machine else 'winding'
    N = machine[wdgk]['num_wires']
    Jmax = 15  # max current density in A/mm2

    i1_max = kwargs.get(
        'i1_max',
        round(0.28*np.pi*hs*(di+hs)/Q1/N*Jmax*1e5)*10 * \
        machine[wdgk].get('num_par_wdgs', 1))

    p = machine['poles']
    num_slices = kwargs.get('num_slices', 3)
    lfe = get_arm_lengths(machine['outer_diam'],
                          machine['inner_diam'],
                          num_slices)
    pole_width = get_pole_widths(machine['outer_diam'],
                                 machine['inner_diam'],
                                 p,
                                 num_slices)

    speed = kwargs.get('speed', 160/p)
    linspeed = [speed*p*pw for pw in pole_width]

    if "num_agnodes" not in machine:
        for pw in pole_width:
            machine['num_agnodes'] = 6*round(pw/machine['airgap']/4)

            #if machine['num_agnodes'] < nper:
            #    machine['num_agnodes'] = 8*round(pw/machine['airgap']/4)

    nlparvardef = {
        "decision_vars": [
            {"values": pole_width,
             "name": "pole_width"},
            {"values": lfe,
             "name": "lfe"},
            {"values": linspeed, "name": "speed"}
        ]
    }
    machine['pole_width'] = np.pi * machine['inner_diam']/machine['poles']
    machine['lfe'] = machine['outer_diam'] - machine['inner_diam']

    beta_min = kwargs.get('beta_min', -90)
    default_steps = round(-beta_min/15)+1
    num_cur_steps = kwargs.get('num_cur_steps', 5)
    num_beta_steps = kwargs.get('num_beta_steps', default_steps)
    parvardef = {
        "decision_vars": [
            #            {"values": sorted(2*temp), "name": "magn_temp"},
            {"steps": num_beta_steps, "bounds": [beta_min, 0], "name": "angl_i_up"},
            {"steps": num_cur_steps,
             "bounds": [i1_max/num_cur_steps, i1_max], "name": "current"}
        ]
    }

    pstudy = parstudy.List(
        workdir, condMat=condMat, magnets=magnetMat,
        magnetizingCurves=magnetizingCurves,
        cmd=kwargs.get('cmd', None))

    ldq = []
    for magtemp in temp:
        nlcalc = dict(
            calculationMode="cogg_calc",
            num_move_steps=60,
            magn_temp=magtemp,
            poc=poc.Poc(999),
            speed=0)
        logging.info("Noload simulation")
        nlresults = pstudy(nlparvardef, machine, nlcalc, engine)
        if nlresults['status'].count('C') != len(nlresults['status']):
            raise ValueError('Noload simulation failed %s', nlresults['status'])
        nlresults.update(process(lfe, pole_width, machine, nlresults['f']))
        current_angles = nlresults['f'][0]['current_angles']

        results = []
        i = 0
        for l, pw in zip(lfe, pole_width):
            mpart = {k: machine[k] for k in machine if k != 'afm_rotor'}
            mpart['pole_width'] = pw
            mpart['lfe'] = l
            subdir = f"{workdir}/{i}"

            gpstudy = parstudy.Grid(
                subdir, condMat=condMat, magnets=magnetMat,
                magnetizingCurves=magnetizingCurves,
                cmd=kwargs.get('cmd', None))

            simulation = dict(
                calculationMode="torq_calc",
                wind_temp=20.0,
                magn_temp=magtemp,
                angl_i_up=0.0,
                magnet_loss=True,
                magn_height=machine['magnet']['afm_rotor']['magn_height'],
                yoke_height=machine['magnet']['afm_rotor'].get(
                    'yoke_height', 0),
                current=1,
                poc=poc.Poc(999,
                            parameters={
                                'phi_voltage_winding': current_angles}),
                num_move_steps=60,
                speed=linspeed[i],
                num_par_wdgs=machine[wdgk].get('num_par_wdgs', 1))

            lresults = gpstudy(parvardef, mpart, simulation, engine)
            f = [{k: bch[k]
                  for k in ('linearForce', 'flux', 'losses', 'lossPar')}
                 for bch in lresults['f']]
            lresults['f'] = f
            results.append(lresults)
            i += 1

        postp = []
        for results in [process(lfe, pole_width, machine, bch)
                        for bch in zip(*[r['f'] for r in results])]:
            torque = np.mean(results.pop('torque'))
            results['torque'] = torque
            results.update(_psidq_ldq(results, nlresults))
            postp.append(results)

        r1 = postp[0]['r1']
        i1 = [r['i1'] for r in postp][::num_beta_steps]
        beta = [r['beta'] for r in postp][:num_beta_steps]
        # Note: Psi RMS Values aligned with BCH/BATCH
        psid = np.reshape([r['psid'] for r in postp],
                          (-1, num_beta_steps)).T/np.sqrt(2)
        psiq = np.reshape([r['psiq'] for r in postp],
                          (-1, num_beta_steps)).T/np.sqrt(2)
        torque = np.reshape([r['torque'] for r in postp],
                            (-1, num_beta_steps)).T
        losses = {k: np.flip(np.reshape([r['plfe'][k] for r in postp],
                                        (-1, num_beta_steps)),
                             axis=1).T.tolist()
                  for k in postp[0]['plfe']}
        losses.update({'speed': speed,
                       'hf': postp[0]['lossPar']['hf'],
                       'ef': postp[0]['lossPar']['ef'],
                       'magnet':np.flip(
                           np.reshape([r['plmag'] for r in postp],
                                      (-1, num_beta_steps)),
                           axis=1).T.tolist()})
        ldq.append({'temperature': magtemp,
                    'i1':i1, 'beta':beta,
                    'psid': psid.tolist(), 'psiq': psiq.tolist(),
                    'torque': torque.tolist(),
                    'losses': losses})
        # T = 3/2 p (Psid iq - Psiq id)
        #iq, id = femagtools.machine.utils.iqd(*np.meshgrid(beta, i1))

    return {'m': machine[wdgk]['num_phases'],
            'p': machine['poles']//2,
            'ls1': 0, 'r1': r1, 'ldq': ldq}


def process(lfe, pole_width, machine, bch):
    """process results: torque, voltage (emf), losses

    Arguments:
    lfe: (float) active machine length
    pole_width: (float) pole width
    machine: (dict) machine
    bch: (list of dict) linearForce, flux, losses

    Returns dict with keys:
    pos: (list of float)
    r1: winding resistance
    torque: (list of float)
    emf: (list of float)
    emf_amp: (float)
    emf_angle: (float)
    freq: (float)
    currents: (list of list of float)
    plfe, plmag, plcu: (list of float)
    """
    model_type = machine['afmtype']
    num_slots = machine['stator']['num_slots']
    mmod = model.MachineModel(machine)
    slots_gen = mmod.stator['num_slots_gen']
    scale_factor =_get_scale_factor(model_type, num_slots, slots_gen)
    endpos = [2*pw*1e3 for pw in pole_width]
    displ = [[d for d in r['linearForce'][0]['displ']
              if d < e*(1+1/len(r['linearForce'][0]['displ']))]
             for e, r in zip(endpos, bch)]
    radius = [pw*machine['poles']/2/np.pi for pw in pole_width]
    rotpos = [np.array(d)/r/1000 for r, d in zip(radius, displ)]
    n = len(rotpos[0])
    currents = [bch[0]['flux'][k][0]['current_k'][:n]
                for k in bch[0]['flux']]
    if len(pole_width) > 1:
        torque = _integrate(radius, rotpos[0], np.array(
            [r*scale_factor*np.array(fx[:-1])/l
             for l, r, fx in zip(lfe, radius,
                                 [r['linearForce'][0]['force_x']
                                  for r in bch])]))

        voltage = {k: [scale_factor * np.array(ux[:-1])/l
                       for l, ux in zip(lfe, [r['flux'][k][0]['voltage_dpsi']
                                              for r in bch])]
                   for k in bch[0]['flux']}
        emf = [_integrate(radius, rotpos[0], np.array(voltage[k]))
               for k in voltage]
    else:
        r = radius[0]
        torque = [r*scale_factor*fx
                  for fx in bch[0]['linearForce'][0]['force_x'][:-1]]
        voltage = {k: [scale_factor * ux
                       for ux in bch[0]['flux'][k][0]['voltage_dpsi'][:-1]]
                   for k in bch[0]['flux']}
        emf = [voltage[k][:n] for k in voltage]

    pos = (rotpos[0]/np.pi*180)
    emffft = utils.fft(pos, emf[0])

    styoke = {}
    stteeth = {'hyst':0, 'eddy':0}
    rotor = {}
    for k in ('hyst', 'eddy'):
        if len(pole_width) > 1:
            styoke[k] = _integrate1d(radius, scale_factor*np.array(
                [sum(b['losses'][0]['stator']['stfe'][k])/l
                 for l, b in zip(lfe, bch)]))
            rotor[k] = _integrate1d(radius, scale_factor*np.array(
                [sum(b['losses'][0]['rotor']['----'][k])/l
                 for l, b in zip(lfe, bch)]))
        else:
            styoke[k] = scale_factor*sum(
                bch[0]['losses'][0]['stator']['stfe'][k])
            rotor[k] = scale_factor*sum(
                bch[0]['losses'][0]['rotor']['----'][k])

    if 'magnetH' in bch[0]['losses'][0]:
        k = 'magnetH'
    else:
        k = 'magnetJ'
    if len(pole_width) > 1:
        maglosses = _integrate1d(radius, scale_factor*np.array(
            [b['losses'][0][k]/l for l, b in zip(lfe, bch)]))
    else:
        maglosses = scale_factor*bch[0]['losses'][0][k]

    freq = bch[0]['losses'][0]['stator']['stfe']['freq'][0]

    wdg = windings.Winding(
        dict(Q=mmod.stator['num_slots'],
             p=mmod.poles//2,
             m=mmod.winding['num_phases'],
             l=mmod.winding['num_layers']))

    cufill = mmod.winding.get('cufilfact', 0.4)
    aw = (mmod.stator['afm_stator']['slot_width']*
              mmod.stator['afm_stator']['slot_height']*
              cufill/mmod.winding['num_wires']/mmod.winding['num_layers'])
    r1 = wdg_resistance(wdg, mmod.winding['num_wires'],
                        mmod.winding['num_par_wdgs'],
                        aw,
                        mmod.outer_diam, mmod.inner_diam)
    i1 = np.mean([np.max(c) for c in currents])/np.sqrt(2)
    plcu = mmod.winding['num_phases']*i1**2*r1

    return {
        'pos': pos.tolist(), 'r1': r1,
        'torque': torque,
        'emf': emf,
        'emf_amp': emffft['a'], 'emf_angle': emffft['alfa0'],
        'freq': freq,
        'currents': currents,
        'beta': bch[0]['losses'][0]['beta'],
        'plfe': {'styoke_hyst': styoke['hyst'], 'styoke_eddy': styoke['eddy'],
                 'stteeth_hyst': stteeth['hyst'], 'stteeth_eddy': stteeth['eddy'],
                 'rotor_hyst': rotor['hyst'], 'rotor_eddy': rotor['eddy']},
        'plmag': maglosses, 'lossPar': {k: bch[0]['lossPar'][k]
                                        for k in bch[0]['lossPar']},
        'plcu': plcu}


def _psidq_ldq(results, nlresults):
    """ calculate currents, winding fluxes psi and inductances L
    Arguments:
    results: (dict) emf, currents, freq of load simulation
    nlresults: (dict) emf of no load simulation

    returns dict of currents, psi, L
    psid, psiq, psim, i1, id, iq, Ld, Lq
    """
    gamma = -(results['emf_angle'] - nlresults['emf_angle'])
    w1 = 2*np.pi*results['freq']
    psid = np.cos(gamma)*results['emf_amp']/w1
    psiq = -np.sin(gamma)*results['emf_amp']/w1
    psim = nlresults['emf_amp']/w1
    i1 = np.mean([np.max(c)
                  for c in results['currents']])/np.sqrt(2)
    beta = results['beta']/180*np.pi
    id = np.sqrt(2)*i1*np.sin(beta)
    iq = np.sqrt(2)*i1*np.cos(beta)
    Ld = Lq = float('nan')
    if np.abs(id) > 0:
        Ld = (psid - psim)/id
    if np.abs(iq) > 0:
        Lq = psiq/iq
    return {
        'psid': psid,
        'psiq': psiq,
        'psim': psim,
        'i1': i1,
        'beta': results['beta'],
        'id': id,
        'iq': iq,
        'Ld': Ld,
        'Lq': Lq}

def _get_scale_factor(model_type, num_slots, slots_gen):
    """Determines the scale factor
    Arguments:
    model_type : (str) type of model
    num_slots: (int) total number of stator slots
    slots_gen: (int) number of slots in model

    Returns scale factor based on number of poles, number of poles simulated
        and the model type
    """
    segments = num_slots / slots_gen
    if model_type in ("S2R1", "S1R2"):
        return 2 * segments
    return segments


def get_arm_lengths(outer_diam, inner_diam, num_slices):
    d = outer_diam - inner_diam
    if num_slices > 2:
        return [d/(4*(num_slices-1))] + [
            d/(2*(num_slices-1))
            for i in range(1,num_slices-1)] + [d/(4*(num_slices-1))]
    return [d/2]


def get_pole_widths(outer_diam, inner_diam, poles, num_slices):
    d = outer_diam - inner_diam
    if num_slices > 2:
        return [np.pi * inner_diam/poles] + [
            np.pi * (inner_diam + d*i/(num_slices - 1))/poles
            for i in range(1, num_slices-1)] + [np.pi * outer_diam/poles]
    return [np.pi * (outer_diam+inner_diam)/2/poles]


def wdg_resistance(wdg, n, g, aw, outer_diam, inner_diam,
                   sigma=56e6):
    """return winding resistance per phase in Ohm
    Arguments:
    wdg: (Winding) winding
    n: (int) number of wires per coil side
    g: (int) number of parallel coil groups
    aw: wire cross section area m2
    outer_diam, inner_diam: diameters m
    sigma: (float) conductivity of wire material 1/Ohm m
    """
    # mean length of one turn
    lt = 2.4*((outer_diam-inner_diam)+np.pi/wdg.Q*(outer_diam+inner_diam) + 16e-3)
    return wdg.turns_per_phase(n, g)*lt/sigma/aw/g


def _get_copper_losses(scale_factor, bch):
    """return copper losses from bch files"""
    try:
        cu_losses = sum([b['losses'][0][wdgk] for b in bch])
        return scale_factor*cu_losses
    except KeyError:
        return 0  # noload calc has no winding losses


class AFPM:
    """Axial Flux PM
    Arguments:
    workdir: (str) working directory
    magnetizingCurves: (str) or (object)
    magnetMat: magnet material
    condMat: conductor material
    """
    def __init__(self, workdir, magnetizingCurves='.', magnetMat='',
                 condMat=''):
        self.parstudy = parstudy.List(
            workdir, condMat=condMat, magnets=magnetMat,
            magnetizingCurves=magnetizingCurves)

    def __call__(self, engine, machine, simulation, num_slices):
        """ run FE simulation

        returns
        pos: (list of float)
        r1: radius
        torque: (list of float)
        emf: (list of float)
        emf_amp: (float)
        emf_angle: (float)
        freq: (float)
        currents: (list of list of float)
        plfe, plmag, plcu: (list of float)
        psid, psiq, psim: (float)
        i1, beta: (float)
        id, iq: (float)
        Ld, Lq: (float)
        """
        try:
            if machine['afmtype'] not in AFM_TYPES:
                raise ValueError(f"invalid afm type {machine['afmtype']}")
        except KeyError:
            raise ValueError("missing key afmtype")

        lfe = get_arm_lengths(machine['outer_diam'],
                              machine['inner_diam'],
                              num_slices)
        pole_width = get_pole_widths(machine['outer_diam'],
                                     machine['inner_diam'],
                                     machine['poles'],
                                     num_slices)
        linspeed = [simulation['speed']*machine['poles']*pw
                    for pw in pole_width]

        if "num_agnodes" not in machine:
            for pw in pole_width:
                machine['num_agnodes'] = 6*round(pw/machine['airgap']/4)
                #Q = machine['stator']['num_slots']
                #p = machine['poles']
                #nper = np.lcm(Q, p)
                #if machine['num_agnodes'] < nper:
                #    machine['num_agnodes'] = 8*round(pw/machine['airgap']/4)

        parvardef = {
            "decision_vars": [
                {"values": pole_width,
                 "name": "pole_width"},
                {"values": lfe,
                 "name": "lfe"},
                {"values": linspeed, "name": "speed"}
            ]
        }
        machine['pole_width'] = np.pi * machine['inner_diam']/machine['poles']
        machine['lfe'] = machine['outer_diam'] - machine['inner_diam']

        nlresults = {}
        if (simulation['calculationMode'] != 'cogg_calc' and
            'poc' not in simulation):
            nlcalc = dict(
                calculationMode="cogg_calc",
                magn_temp=simulation.get('magn_temp', 20),
                num_move_steps=60,
                poc=poc.Poc(machine['pole_width']),
                speed=0)
            logging.info("Noload simulation")
            nlresults = self.parstudy(parvardef,
                                      machine, nlcalc, engine)
            if nlresults['status'].count('C') != len(nlresults['status']):
                raise ValueError('Noload simulation failed %s', nlresults['status'])
            nlresults.update(process(lfe, pole_width, machine, nlresults['f']))

            current_angles = nlresults['f'][0]['current_angles']
            simulation['poc'] = poc.Poc(
                999,
                parameters={
                    'phi_voltage_winding': current_angles})
            logger.info("Current angles: %s", current_angles)

        lresults = self.parstudy(
            parvardef,
            {k: machine[k] for k in machine if k!= 'afm_rotor'},
            simulation, engine)  # Note: imcomplete machine prevents rebuild

        results = process(lfe, pole_width, machine, lresults['f'])
        results.update(_psidq_ldq(results, nlresults))
        return results
